/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Application.h"

#include "About.h"				// DoAbout
#include "Byteswapping.h"		// Canonical
#include "CPU_MT.h" 			// CPUStopper
#include "DebugMgr.h"			// BreakpointCondition, etc.
#include "Document.h"			// QueryNewDocument, QueryOpenDocument, gCurrentRAMPath
#include "EmDlg.h"				// DoEditLoggingOptions
#include "EmDlgWin.h"			// CenterWindow, CenterDlgProc
#include "EmROMTransfer.h"		// EmROMTransfer::ROMTransfer
#include "EmStreamFile.h"		// EmStreamFile, kCreateAlways
#include "Emulator.h"			// gInstance
#include "EmulatorTypes.h"		// CloseActionType
#include "Hordes.h"				// Hordes::New, etc.
#include "HostPreferences.h"	// gHostPrefs
#include "LCD.h"				// LCD_xxx functions
#include "LoadApplication.h"	// LoadPalmFileList
#include "Miscellaneous.h"		// GetApplicationNames
#include "MMFMessaging.h"		// MMFOpenFiles, MMFCloseFiles
#include "Platform.h"			// Platform::GetString
#include "PreferenceMgr.h"		// Preference
#include "Profiling.h"			// ProfileInit, etc.
#include "EmRPC.h"				// RPC::Idle
#include "Skins.h"				// kElement_PowerButton, etc.
#include "SocketMessaging.h"	// CSocket::IdleAll
#include "Sounds.h"
#include "Startup.h"			// ScheduleCreateNewSession, etc.
#include "Strings.r.h"			// kStr_ values

#include "resource.h"

#include <tchar.h>				// _tcsinc, etc.


// ======================================================================
//	Globals and constants
// ======================================================================

typedef void	(*MenuHandler) (HWND hwnd, WPARAM command);
typedef Bool	(*UpdateHandler) (HWND hwnd, WPARAM command);

WORD			gWinSockVersion;
Bool			gOpenGCWWindow;

static DWORD	gHotSyncDownStart;

static WNDPROC	g_lpfnDialogProc;


// ======================================================================
//	Private functions
// ======================================================================

typedef char* POSITION;

static POSITION 		GetStartPosition	(OPENFILENAME& ofn);
static string			GetNextPathName 	(OPENFILENAME& ofn, POSITION& pos);

static UINT CALLBACK	OpenMultiFileProc	(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK OpenMultiFileHook	(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

static void	PrvLoadPalmFiles	(EmFileRefList& fileList);
static Bool PrvCloseSession		(HWND hwnd, const EmFileRef& saveFile, CloseActionType action);
static Bool PrvCloseSession		(HWND hwnd);

static void DoAbout					(HWND hwnd, WPARAM command);
	   void DoNew					(HWND hwnd, WPARAM command);
static void DoOpenOther				(HWND hwnd, WPARAM command);
	   void DoOpenRecent			(HWND hwnd, WPARAM command);
static void DoClose					(HWND hwnd, WPARAM command);
static void DoSave					(HWND hwnd, WPARAM command);
static void DoSaveAs				(HWND hwnd, WPARAM command);
static void DoSaveBound				(HWND hwnd, WPARAM command);
static void DoSaveScreen			(HWND hwnd, WPARAM command);
static void DoInstallOther			(HWND hwnd, WPARAM command);
static void DoInstallRecent			(HWND hwnd, WPARAM command);
static void DoExportDatabase		(HWND hwnd, WPARAM command);
static void DoHotSync				(HWND hwnd, WPARAM command);
static void DoReset					(HWND hwnd, WPARAM command);
static void DoTransferROM			(HWND hwnd, WPARAM command);
static void DoGremlinsNew			(HWND hwnd, WPARAM command);
	   void DoGremlinsStep			(HWND hwnd, WPARAM command);
	   void DoGremlinsResume		(HWND hwnd, WPARAM command);
	   void DoGremlinsStop			(HWND hwnd, WPARAM command);
static void DoHordesFromControlFile	(HWND hwnd, WPARAM command);
static void DoProfileStart			(HWND hwnd, WPARAM command);
static void DoProfileStop			(HWND hwnd, WPARAM command);
static void DoProfileDump			(HWND hwnd, WPARAM command);
static void DoProperties			(HWND hwnd, WPARAM command);
static void DoLoggingOptions		(HWND hwnd, WPARAM command);
static void DoDebugOptions			(HWND hwnd, WPARAM command);
static void DoTracingOptions		(HWND hwnd, WPARAM command);
static void DoSkins					(HWND hwnd, WPARAM command);
static void DoBreakpoints			(HWND hwnd, WPARAM command);
static void DoMinimize				(HWND hwnd, WPARAM command);
static void DoExit					(HWND hwnd, WPARAM command);


static const struct
{
	int				fCommand;
	int 			fOperation;
	MenuHandler 	fHandler;
	UpdateHandler	fUpdater;
	int				fStopType;
}
kCommandHandlers[] =
{
	{ ID_EXIT,					kStr_CmdQuit,			DoExit, 			NULL,	0 },

	{ ID_NEW, 					kStr_CmdNew,			DoNew,				NULL,	0 },
	{ ID_OPEN_OTHER,			kStr_CmdOpen,			DoOpenOther,		NULL,	0 },
	{ ID_OPEN_RECENT + 0, 		kStr_CmdOpen,			DoOpenRecent,		NULL,	0 },
	{ ID_OPEN_RECENT + 1, 		kStr_CmdOpen,			DoOpenRecent,		NULL,	0 },
	{ ID_OPEN_RECENT + 2, 		kStr_CmdOpen,			DoOpenRecent,		NULL,	0 },
	{ ID_OPEN_RECENT + 3, 		kStr_CmdOpen,			DoOpenRecent,		NULL,	0 },
	{ ID_OPEN_RECENT + 4, 		kStr_CmdOpen,			DoOpenRecent,		NULL,	0 },
	{ ID_OPEN_RECENT + 5, 		kStr_CmdOpen,			DoOpenRecent,		NULL,	0 },
	{ ID_OPEN_RECENT + 6, 		kStr_CmdOpen,			DoOpenRecent,		NULL,	0 },
	{ ID_OPEN_RECENT + 7, 		kStr_CmdOpen,			DoOpenRecent,		NULL,	0 },
	{ ID_OPEN_RECENT + 8,		kStr_CmdOpen,			DoOpenRecent,		NULL,	0 },
	{ ID_OPEN_RECENT + 9,		kStr_CmdOpen,			DoOpenRecent,		NULL,	0 },
	{ ID_CLOSE,					kStr_CmdClose,			DoClose,			NULL,	0 },

	{ ID_SAVE,					kStr_CmdSave,			DoSave, 			NULL,	kStopOnADime },
	{ ID_SAVEAS,				kStr_CmdSave,			DoSaveAs,			NULL,	kStopOnADime },
	{ ID_SAVEBOUND,				kStr_CmdSaveBound,		DoSaveBound,		NULL,	kStopOnADime },
	{ ID_SAVESCREEN,			kStr_CmdSaveScreen, 	DoSaveScreen,		NULL,	kStopOnADime },

	{ ID_INSTALL_OTHER,			kStr_CmdInstall,		DoInstallOther, 	NULL,	kStopOnATrap },
	{ ID_INSTALL_RECENT + 0,	kStr_CmdInstall,		DoInstallRecent,	NULL,	kStopOnATrap },
	{ ID_INSTALL_RECENT + 1,	kStr_CmdInstall,		DoInstallRecent,	NULL,	kStopOnATrap },
	{ ID_INSTALL_RECENT + 2,	kStr_CmdInstall,		DoInstallRecent,	NULL,	kStopOnATrap },
	{ ID_INSTALL_RECENT + 3,	kStr_CmdInstall,		DoInstallRecent,	NULL,	kStopOnATrap },
	{ ID_INSTALL_RECENT + 4,	kStr_CmdInstall,		DoInstallRecent,	NULL,	kStopOnATrap },
	{ ID_INSTALL_RECENT + 5,	kStr_CmdInstall,		DoInstallRecent,	NULL,	kStopOnATrap },
	{ ID_INSTALL_RECENT + 6,	kStr_CmdInstall,		DoInstallRecent,	NULL,	kStopOnATrap },
	{ ID_INSTALL_RECENT + 7,	kStr_CmdInstall,		DoInstallRecent,	NULL,	kStopOnATrap },
	{ ID_INSTALL_RECENT + 8,	kStr_CmdInstall,		DoInstallRecent,	NULL,	kStopOnATrap },
	{ ID_INSTALL_RECENT + 9,	kStr_CmdInstall,		DoInstallRecent,	NULL,	kStopOnATrap },
	{ ID_EXPORT_DATABASE, 		kStr_CmdExportDatabase,	DoExportDatabase,	NULL,	kStopOnATrap },
	{ ID_HOTSYNC, 				kStr_CmdHotSync,		DoHotSync,			NULL,	kStopOnADime },
	{ ID_RESET,					kStr_CmdReset,			DoReset,			NULL,	0 },
	{ ID_TRANSFERROM, 			kStr_CmdTransferROM,	DoTransferROM,		NULL,	0 },

	{ ID_GREMLINS_NEW,			kStr_CmdGremlinNew, 	DoGremlinsNew,		NULL,	kStopOnATrap },
	{ ID_GREMLINS_STEP,			kStr_CmdGremlinStep,	DoGremlinsStep, 	NULL,	kStopOnATrap },
	{ ID_GREMLINS_RESUME, 		kStr_CmdGremlinResume,	DoGremlinsResume,	NULL,	kStopOnATrap },
	{ ID_GREMLINS_STOP,			kStr_CmdGremlinStop,	DoGremlinsStop,		NULL,	kStopOnATrap },

	// TODO: create string constant for operation description

	{ ID_HORDES_FROM_CONTROL_FILE, kStr_CmdGremlinResume, DoHordesFromControlFile, NULL,	kStopOnADime },

#if HAS_PROFILING
	{ ID_PROFILE_START,			kStr_CmdProfileStart,	DoProfileStart, 	NULL,	kStopOnADime },
	{ ID_PROFILE_STOP,			kStr_CmdProfileStop,	DoProfileStop,		NULL,	kStopOnADime },
	{ ID_PROFILE_DUMP,			kStr_CmdProfileDump,	DoProfileDump,		NULL,	kStopOnADime },
#endif

	{ ID_PROPERTIES,			kStr_CmdPreferences,	DoProperties,		NULL,	kStopOnATrap },
	{ ID_LOGGINGOPTIONS,		kStr_CmdLoggingOptions, DoLoggingOptions,	NULL,	kStopOnADime },
	{ ID_DEBUGOPTIONS,			kStr_CmdDebugOptions,	DoDebugOptions, 	NULL,	kStopOnADime },
	{ ID_TRACINGOPTIONS,		kStr_CmdTracingOptions,	DoTracingOptions, 	NULL,	kStopOnADime },
	{ ID_SKINS,					kStr_CmdSkins,			DoSkins,		 	NULL,	kStopOnADime },
	{ ID_BREAKPOINTS, 			kStr_CmdBreakpoints,	DoBreakpoints,		NULL,	kStopOnADime },

//	{ ID_MINIMIZE,				kStr_CmdMinimize,		DoMinimize,			NULL,	0 },
	{ ID_ABOUT,					kStr_CmdAbout,			DoAbout,			NULL,	0 },

	{ NULL,						kStr_GenericOperation,	NULL,				NULL,	0 }
};


/***********************************************************************
 *
 * FUNCTION:	PrvValidateNumber
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

static int PrvValidateNumber(	const char* text,
								long low,
								long high,
								const char* label,
								long* result)
{
	// !!! Localize calls to this function.

	const char *nptr = text;
	int c;

	while ( isspace((int)(unsigned char)*nptr) )
		++nptr;

	c = (int)(unsigned char) *nptr++;
	int sign = c;
	if (c == '-' || c == '+')
	{
		c = (int)(unsigned char) *nptr++;
	}

	*result = 0;

	while (isdigit(c))
	{
		*result = 10 * *result + (c - '0');
		c = (int)(unsigned char) *nptr++;
	}

	if (sign == '-')
	{
		*result *= -1;
	}

	if (*result < low || *result > high)
	{
		char	buffer[200];
		// !!! Put in resource; use Errors::CommonDialog
		sprintf (buffer, "The value entered for %s must be in the range %ld to %ld.",
			label, low, high);
		::MessageBox (NULL, buffer, "Error", MB_OK);
		
		return false;
	}

	if (nptr <= text + strlen (text))
	{
		char	buffer[200];
		// !!! Put in resource; use Errors::CommonDialog
		sprintf (buffer, "A non-numeric character was entered for %s.", label);
		::MessageBox (NULL, buffer, "Error", MB_OK);

		return false;
	}

	return true;
}

/***********************************************************************
 *
 * FUNCTION:	GetStartPosition
 *
 * DESCRIPTION:	Iterator-style function returning the starting
 *				position for the list of files returned from
 *				GetOpenFileName.  Call this function first; then call
 *				GetNextPathName for as long as POSITION is non-zero.
 *
 *				Function unabashedly stolen from MFC.
 *
 * PARAMETERS:	ofn - the OPENFILENAME passed to GetOpenFileName.
 *
 * RETURNED:	An opaque iterator type containing a value indicating
 *				the first item in the list.
 *
 ***********************************************************************/

static POSITION GetStartPosition (OPENFILENAME& ofn)
{
	return (POSITION) ofn.lpstrFile;
}


/***********************************************************************
 *
 * FUNCTION:	GetNextPathName
 *
 * DESCRIPTION:	Get the pathname indicated by pos, and then increment
 *				pos to refer to the next file in the list.
 *
 *				Function unabashedly stolen from MFC.
 *
 * PARAMETERS:	ofn - the OPENFILENAME passed to GetOpenFileName.
 *
 *				pos - POSITION originally obtained by calling
 *					GetStartPosition.  Subsequent values are returned
 *					by updating pos in-place.
 *
 * RETURNED:	The next pathname indicated by pos.
 *
 ***********************************************************************/

static string GetNextPathName (OPENFILENAME& ofn, POSITION& pos)
{
	BOOL	bExplorer	= (ofn.Flags & OFN_EXPLORER) != 0;
	TCHAR	chDelimiter = bExplorer ? '\0' : ' ';
	LPTSTR	lpsz		= (LPTSTR) pos;

	if (lpsz == ofn.lpstrFile) // first time
	{
		if ((ofn.Flags & OFN_ALLOWMULTISELECT) == 0)
		{
			pos = NULL;
			return ofn.lpstrFile;
		}

		// find char pos after first Delimiter
		while (*lpsz != chDelimiter && *lpsz != '\0')
		{
			lpsz = _tcsinc (lpsz);
		}

		lpsz = _tcsinc (lpsz);

		// if single selection then return only selection
		if (*lpsz == 0)
		{
			pos = NULL;
			return ofn.lpstrFile;
		}
	}

	string strPath = ofn.lpstrFile;
	if (!bExplorer)
	{
		LPTSTR	lpszPath = ofn.lpstrFile;

		while (*lpszPath != chDelimiter)
		{
			lpszPath = _tcsinc (lpszPath);
		}

		strPath = strPath.substr (0, lpszPath - ofn.lpstrFile);
	}

	LPTSTR	lpszFileName = lpsz;
	string	strFileName = lpsz;

	// find char pos at next Delimiter
	while (*lpsz != chDelimiter && *lpsz != '\0')
	{
		lpsz = _tcsinc(lpsz);
	}

	if (!bExplorer && *lpsz == '\0')
	{
		pos = NULL;
	}
	else
	{
		if (!bExplorer)
		{
			strFileName = strFileName.substr (0, lpsz - lpszFileName);
		}

		lpsz = _tcsinc (lpsz);

		if (*lpsz == '\0') // if double terminated then done
			pos = NULL;
		else
			pos = (POSITION) lpsz;
	}

	// only add '\\' if it is needed
	if (!strPath.empty())
	{
		// check for last back-slash or forward slash (handles DBCS)
		LPCTSTR lpsz = _tcsrchr (strPath.c_str (), '\\');
		if (lpsz == NULL)
			lpsz = _tcsrchr (strPath.c_str (), '/');

		// if it is also the last character, then we don't need an extra
		if (lpsz != NULL &&
			(lpsz - strPath.c_str ()) == strPath.size () - 1)
		{
			assert (*lpsz == '\\' || *lpsz == '/');
			return strPath + strFileName;
		}
	}

	return strPath + '\\' + strFileName;
}


/***********************************************************************
 *
 * FUNCTION:	OpenMultiFileProc
 *
 * DESCRIPTION:	Hook the OpenFileName dialog so that we can look for
 *				presses on the OK button.  We look for those so that
 *				we can ensure that the buffer receiving the selection
 *				is large enough.  From MSDN:
 *
 *				"With the introduction of the new common dialogs for
 *				Windows 95, a new way of handling the FNERR_BUFFERTOOSMALL
 *				error was developed. It is still necessary to watch for
 *				the Open button to be pressed and reallocate the buffer
 *				if needed, but the way to watch for the OK is much different. 
 *
 *				"When you install a hook on the Open File common dialog
 *				in Windows 95 using the OPENFILENAME.lpfnHook member,
 *				the dialog you are hooking is a child of the main Open
 *				File dialog. Therefore, to intercept the OK button, you
 *				need to subclass the parent dialog. To do this, you can
 *				install the hook procedure and watch for the CDN_INITDONE
 *				notification. The Open File dialog will send this as part
 *				of a WM_NOTIFY message when the initialization for the
 *				dialog is complete."
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

static UINT CALLBACK OpenMultiFileProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	UNUSED_PARAM(wParam);

	static HWND hwndParentDialog;
	LPOFNOTIFY lpofn;

	switch (uMsg)
	{
		case WM_INITDIALOG:
			// You need to use a copy of the OPENFILENAME struct used to
			// create this dialog. You can store a pointer to the
			// OPENFILENAME struct in the ofn.lCustData so you can retrieve
			// it here in the lParam. Once you have it, you need to hang on
			// to it. Using window properties provides a good thread safe
			// solution to using a global variable.

			::SetProp(hwnd, "OFN", (HANDLE) lParam);
			::CenterWindow (::GetParent (hwnd));
			return (0);

		case WM_NOTIFY:
			// The OFNOTIFY struct is passed in the lParam of this message.

			lpofn = (LPOFNOTIFY) lParam;

			switch (lpofn->hdr.code)
			{
				case CDN_INITDONE:
					// Subclass the parent dialog to watch for the OK
					// button.

					hwndParentDialog = ::GetParent(hwnd);
					g_lpfnDialogProc = (WNDPROC) ::SetWindowLong(hwndParentDialog,
											DWL_DLGPROC, (LONG) &::OpenMultiFileHook);
					break;

			}
			return (0);

		case WM_DESTROY:
			// Need to clean up the subclassing we did on the dialog.
			::SetWindowLong(hwndParentDialog, DWL_DLGPROC, (LONG) g_lpfnDialogProc);

			// Also need to free the property with the OPENFILENAME struct
			::RemoveProp(hwnd, "OFN");
			return (0);
	}
	return (0);
}


/***********************************************************************
 *
 * FUNCTION:	OpenMultiFileHook
 *
 * DESCRIPTION:	Part 2 of hooking the Open File dialog so that we can
 *				look for clicks on OK.	See OpenMultiFileProc for the
 *				discussion of the first part.  Continuing the MSDN
 *				description:
 *
 *				"Once the parent dialog is subclassed, the program can
 *				watch for the actual Open button. When the program gets
 *				the Open button command, it needs to check to see if the
 *				buffer originally allocated is large enough to handle
 *				all the files selected. The CommDlg_OpenSave_GetFilePath()
 *				API will return the length needed."
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

static LRESULT CALLBACK OpenMultiFileHook(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	LPTSTR			lpsz;
	int 			cbLength;
	LPOPENFILENAME	lpofn;

	switch (uMsg)
	{
		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDOK:
					// Need to verify the original buffer size is large
					// enough to handle the files selected. The
					// CommDlg_OpenSave_GetFilePath() API will return the
					// length needed for this buffer.

					cbLength = CommDlg_OpenSave_GetFilePath(hwnd, NULL, 0);

					// MAX_PATH is the size of the buffer originally
					// used in the OPENFILENAME.lpstrFile member.

					if (MAX_PATH < cbLength)
					{
						// The buffer is too small, so allocate a
						// new buffer.
						lpsz = (LPTSTR) malloc (cbLength);
						if (lpsz)
						{
							// The OFN struct is stored in a property of
							// the dialog window.

							lpofn = (LPOPENFILENAME) ::GetProp(hwnd, "OFN");

							lpofn->lpstrFile = lpsz;
							lpofn->nMaxFile  = cbLength;
						}
					}

					// Now let the dialog handle the message normally.
					break;
			}
			break;
	}

	return (::CallWindowProc(g_lpfnDialogProc, hwnd, uMsg, wParam, lParam));
}


/***********************************************************************
 *
 * FUNCTION:	ExportDatabaseProc
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

BOOL CALLBACK ExportDatabaseProc (	HWND hwnd,
									UINT msg,
									WPARAM wparam,
									LPARAM lparam)
{
	static DatabaseInfo*		result;
	static DatabaseInfoList		dbList;

	switch (msg)
	{
		default:
			return FALSE;	// The dialog box procedure should return nonzero if it processes the message, and zero if it does not.

		case WM_DESTROY:
			dbList.clear ();
			break;

		case WM_COMMAND:
			switch (LOWORD (wparam))
			{
				default:
					return FALSE;	// The dialog box procedure should return nonzero if it processes the message, and zero if it does not.

				case IDOK:
				{
					HWND	list = GetDlgItem (hwnd, IDC_APPLICATION_LIST);
					LRESULT itemIndex = ::SendMessage (list, LB_GETCURSEL, 0, 0);

					if (itemIndex != LB_ERR)
					{
						*result = dbList [itemIndex];
					}
					else
					{
						// Treat no selection like Cancel.
						wparam = IDCANCEL;
					}

					// *** FALL THROUGH ***
				}

				case IDCANCEL:
				{
					::EndDialog (hwnd, LOWORD (wparam));
					break;
				}
			}
			break;

		case WM_INITDIALOG:
		{
			result = (DatabaseInfo*) lparam;

			::CenterWindow (hwnd);

			HWND	list = GetDlgItem (hwnd, IDC_APPLICATION_LIST);
			Bool	selected = false;

			// Get the list of installed databases.
			// (I tried doing this in WM_CREATE, but that's not passed to a DlgProc).

			::GetDatabases (dbList, kAllDatabases);

			for (DatabaseInfoList::size_type ii = 0; ii < dbList.size (); ++ii)
			{
				char tempNameBuf[dmDBNameLength*2+4];

				strcpy(tempNameBuf, dbList[ii].dbName);
				if (strcmp(dbList[ii].dbName, dbList[ii].name))
					sprintf(tempNameBuf+strlen(tempNameBuf), " (%s)", dbList[ii].name);

				::SendMessage (list, LB_ADDSTRING, 0, (LPARAM) tempNameBuf);
			}

			/*
				In response to a WM_INITDIALOG message, the dialog box procedure should
				return zero if it calls the SetFocus function to set the focus to one of
				the controls in the dialog box. Otherwise, it should return nonzero, in
				which case the system sets the focus to the first control in the dialog
				box that can be given the focus.
			*/

			break;
		}
	}

	return TRUE;	// The dialog box procedure should return nonzero if it processes the message, and zero if it does not.
}



// Parse an integer, which may be either decimal ("12345") or hex ("0x12345").
// Returns true if OK, false on parse error.
BOOL parseInt(char *s, UInt32 *i)
{
	if (sscanf(s, " %i", i) == 1)
		return TRUE;

	return FALSE;
}

static BOOL CALLBACK CodeBreakpointProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch (msg)
	{
		case WM_COMMAND:
			switch (LOWORD (wparam))
			{
				case IDOK:
				case IDCANCEL:
					if (LOWORD (wparam) == IDOK)
					{
						char	text[256];
						UInt32	address;

						::GetDlgItemText (hwnd, IDC_ADDRESS, text, sizeof (text));
						if (!parseInt (text, &address))
						{
							::MessageBox (NULL, "bad address", NULL, MB_ICONEXCLAMATION);
							break;
						}

						::GetDlgItemText (hwnd, IDC_CONDITION, text, sizeof text);
						BreakpointCondition* c = NULL;

						if (*text)
						{
							c = Debug::NewBreakpointCondition(text);
							if (!c)
							{
								::MessageBox (NULL, "can't parse condition", NULL, MB_ICONEXCLAMATION);
								break;
							}
						}

						// install
						int i = ::GetWindowLong (hwnd, DWL_USER);
						Debug::SetBreakpoint (i, address, c);
					}

					EndDialog (hwnd, LOWORD (wparam));
					break;
			}
			break;

		case WM_INITDIALOG:
		{
			char text[256];
			int i = lparam; 	// index of code breakpoint to edit

			::CenterWindow (hwnd);

			if (gDebuggerGlobals.bp[i].enabled)
			{
				sprintf (text, "0x%x", gDebuggerGlobals.bp[i].addr);
				::SetDlgItemText (hwnd, IDC_ADDRESS, text);

				BreakpointCondition* c = gDebuggerGlobals.bpCondition[i];
				if (c)
					::SetDlgItemText (hwnd, IDC_CONDITION, c->source);
			}

			::SetWindowLong (hwnd, DWL_USER, i);
			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}

static void CodeBreakpointDialog (HWND hwnd, int i)
// Edit the code breakpoint at index i
{
	::DialogBoxParam (gInstance, MAKEINTRESOURCE (IDD_CODEBREAKPOINT), hwnd,
		&::CodeBreakpointProc, i);
}

static void enableControl (HWND hwnd, int control, bool enable)
{
	HWND w = ::GetDlgItem (hwnd, control);
	::EnableWindow (w, enable);
}

static void EnableCodeBreakpointControls (HWND hwnd, bool enable)
{
	::enableControl (hwnd, IDC_EDIT, enable);
	::enableControl (hwnd, IDC_CLEAR, enable);
}

static void EnableDataBreakpointControls (HWND hwnd, bool enable)
{
	::enableControl (hwnd, IDC_STARTADDRESS, enable);
	::enableControl (hwnd, IDC_ADDRESS, enable);
	::enableControl (hwnd, IDC_NUMBEROFBYTES, enable);
	::enableControl (hwnd, IDC_NBYTES, enable);
}

// return the index of the breakpoint which is currently selected, or LB_ERR if none
static int selectedBreakpoint (HWND hwnd)
{
	HWND list = ::GetDlgItem (hwnd, IDC_LIST);
	return ListBox_GetCurSel (list);
}

static void RefreshBreakpointList (HWND hwnd)
{
	HWND list = ::GetDlgItem (hwnd, IDC_LIST);
	SetWindowRedraw (list, false);

	int selected = ListBox_GetCurSel (list);
	ListBox_ResetContent (list);

	for (int i = 0 ; i < dbgTotalBreakpoints ; ++i)
	{
		char text[256];
		if (gDebuggerGlobals.bp[i].enabled)
		{
			sprintf(text, "at 0x%x", gDebuggerGlobals.bp[i].addr);
			BreakpointCondition* c = gDebuggerGlobals.bpCondition[i];
			if (c)
			{
				strcat (text, " when ");
				strcat (text, c->source);
			}
		}
		else
		{
			sprintf (text, "(#%d - not set)", i);
		}

		ListBox_AddString (list, text);
	}

	if (selected != LB_ERR)
		ListBox_SetCurSel (list, selected);

	SetWindowRedraw (list, true);
	::InvalidateRect (list, NULL, true);
}

static bool intToBool (int i)
// convert an int to a bool without generating a compiler warning
{
	return i != 0;
}

static BOOL CALLBACK BreakpointsProc (HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	UNUSED_PARAM(lparam);

	switch (msg)
	{
		case WM_COMMAND:
			switch (LOWORD (wparam))
			{
				case IDOK:
				// even though we don't have a Cancel button, the dialog has an X box which
				// sends an IDCANCEL notification
				case IDCANCEL:	
				{
					// set data breakpoint
					char	text[256];
					bool	watchEnabled;
					UInt32	watchAddr, watchBytes;

					HWND	check = ::GetDlgItem (hwnd, IDC_ENABLED);

					watchEnabled = intToBool(Button_GetCheck(check));
					if (watchEnabled)
					{
						::GetDlgItemText (hwnd, IDC_ADDRESS, text, sizeof (text));
						if (!parseInt (text, &watchAddr))
						{
							::MessageBox (NULL, "bad address", NULL, MB_ICONEXCLAMATION);
							break;
						}

						::GetDlgItemText (hwnd, IDC_NBYTES, text, sizeof (text));
						if (!parseInt (text, &watchBytes))
						{
							::MessageBox (NULL, "bad number of bytes", NULL, MB_ICONEXCLAMATION);
							break;
						}

						gDebuggerGlobals.watchAddr = watchAddr;
						gDebuggerGlobals.watchBytes = watchBytes;
					}

					gDebuggerGlobals.watchEnabled = watchEnabled;

					::EndDialog (hwnd, LOWORD (wparam));
					break;
				}

				case IDC_LIST:
				{
					WORD	code = HIWORD(wparam);
					if (code == LBN_SELCHANGE)	// a list item was selected
					{
						int index = ::selectedBreakpoint (hwnd);
						::EnableCodeBreakpointControls (hwnd, index != LB_ERR);
					}

					if (code != LBN_DBLCLK)
						break;
				}
					/* FALL THROUGH */
				case IDC_EDIT:
				{
					int index = ::selectedBreakpoint (hwnd);
					if (index != LB_ERR)
					{
						::CodeBreakpointDialog (hwnd, index);
						::RefreshBreakpointList (hwnd);
					}
					break;
				}

				case IDC_CLEAR:
				{
					int index = ::selectedBreakpoint (hwnd);
					if (index != LB_ERR)
					{
						Debug::ClearBreakpoint (index);
						::RefreshBreakpointList (hwnd);
					}
					break;
				}

				case IDC_ENABLED:
				{
					HWND	check = ::GetDlgItem (hwnd, IDC_ENABLED);
					bool	c = ::intToBool (Button_GetCheck (check));
					::EnableDataBreakpointControls (hwnd, c);
					break;
				}
			}
			break;

		case WM_INITDIALOG:
		{
			::CenterWindow (hwnd);

			char text[256];

			// code breakpoints
			::RefreshBreakpointList (hwnd);

			// Initially no breakpoint is selected in the list, so the "edit" and
			// "clear" buttons should be disabled.
			::EnableCodeBreakpointControls (hwnd, false);

			// data breakpoint

			HWND	check = ::GetDlgItem (hwnd, IDC_ENABLED);
			Button_SetCheck (check, gDebuggerGlobals.watchEnabled);

			HWND	address = ::GetDlgItem (hwnd, IDC_ADDRESS);
			sprintf (text, "0x%x", gDebuggerGlobals.watchAddr);
			Edit_SetText (address, text);

			HWND	bytes = ::GetDlgItem (hwnd, IDC_NBYTES);
			sprintf (text, "0x%x", gDebuggerGlobals.watchBytes);
			Edit_SetText (bytes, text);

			::EnableDataBreakpointControls (hwnd, gDebuggerGlobals.watchEnabled);

			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}

static void DoBreakpointsDialog (HWND hwnd)
{
	::DialogBox (gInstance, MAKEINTRESOURCE (IDD_BREAKPOINTS), hwnd, &::BreakpointsProc);
}


/***********************************************************************
 *
 * FUNCTION:	PrvLoadPalmFiles
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void PrvLoadPalmFiles (EmFileRefList& fileList)
{
	::LoadPalmFileList (fileList);
}



/***********************************************************************
 *
 * FUNCTION:	PrvCloseSession
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	True if the session was closed, false if the user canceled.
 *
 ***********************************************************************/

Bool PrvCloseSession (HWND hwnd, const EmFileRef& saveFile, CloseActionType action)
{
	// If we're to save the session, then save it.

	if (action == kSaveAlways)
	{
		// Stop the CPU thread.  It will get stopped in our WM_DESTROY handler
		// anyway, but we need to stop it here in order to save the RAM image.

		CPUStopper	stopper (kStopOnADime);
		if (!stopper.Stopped ())
			return false;

		if (!::SaveRAMImage (hwnd, saveFile, !kAskUser, !kForAutoSave))
			return false;
	}

	// Close the LCD window (leading the UI thread to quit).

	::DestroyWindow (hwnd);

	return true;
}

Bool PrvCloseSession (HWND hwnd)
{
	// Get the user's preference as to what we're supposed to do.

	Preference<CloseActionType> pref1 (kPrefKeyCloseAction);
	CloseActionType 			action = *pref1;

	if (::IsBoundFully ())
	{
		action = kSaveNever;
	}

	// Get the name of the last-saved session file.

	Preference<EmFileRef>	pref2 (kPrefKeyLastPSF);
	EmFileRef				saveFile = *pref2;

	// If the user wants us to ask before quitting, then ask.

	if (action == kSaveAsk)
	{
		char	buffer[300];
		string	saveChanges (Platform::GetString (IDS_SAVECHANGES));
		string	untitled (Platform::GetString (kStr_Untitled));
		string	app (Platform::GetString (kStr_AppName));

		sprintf (buffer, saveChanges.c_str (), saveFile.IsSpecified ()
										? saveFile.GetName ().c_str ()
										: untitled.c_str ());
		int result = ::MessageBox (hwnd, buffer, app.c_str (),
						MB_APPLMODAL | MB_ICONQUESTION | MB_YESNOCANCEL);

		if (result == IDYES)
			action = kSaveAlways;

		else if (result == IDNO)
			action = kSaveNever;

		else		// result == IDCANCEL
			return false;
	}

	return PrvCloseSession (hwnd, saveFile, action);
}


/***********************************************************************
 *
 * FUNCTION:	DoAbout
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoAbout (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	::DoAbout (hwnd);
}


/***********************************************************************
 *
 * FUNCTION:	DoNew
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoNew (HWND hwnd, WPARAM command)
{
	Preference<Configuration>	pref (kPrefKeyLastConfiguration);
	Configuration	cfg = *pref;

	if (::QueryNewDocument (hwnd, cfg))
	{
		if (::PrvCloseSession (hwnd))
		{
			Startup::ScheduleCreateSession(cfg);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoOpenOther
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoOpenOther (HWND hwnd, WPARAM command)
{
	EmFileRef	ref;

	if (::QueryOpenDocument (hwnd, ref))
	{
		if (::PrvCloseSession (hwnd))
		{
			Startup::ScheduleOpenSession(ref);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoOpenRecent
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoOpenRecent (HWND hwnd, WPARAM command)
{
	if (command >= ID_OPEN_RECENT &&
		command < ID_OPEN_RECENT + MRU_COUNT)
	{
		if (::PrvCloseSession (hwnd))
		{
			EmFileRef	ref (gEmuPrefs->GetIndRAMMRU (command - ID_OPEN_RECENT));
			Startup::ScheduleOpenSession (ref);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoClose
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoClose (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	if (::PrvCloseSession (hwnd))
	{
		Startup::ScheduleAskWhatToDo();
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoSave
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoSave (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	Preference<EmFileRef>	pref (kPrefKeyLastPSF);
	EmFileRef				currentRAMRef = *pref;
	::SaveRAMImage (hwnd, currentRAMRef, !kAskUser, !kForAutoSave);
}


/***********************************************************************
 *
 * FUNCTION:	DoSaveAs
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoSaveAs (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	Preference<EmFileRef>	pref (kPrefKeyLastPSF);
	EmFileRef				currentRAMRef = *pref;
	::SaveRAMImage (hwnd, currentRAMRef, kAskUser, !kForAutoSave);
}


/***********************************************************************
 *
 * FUNCTION:	DoSaveBound
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoSaveBound (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(hwnd);
	UNUSED_PARAM(command);

	EmDlg::DoSaveBound ();
}


/***********************************************************************
 *
 * FUNCTION:	DoSaveScreen
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoSaveScreen (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(hwnd);
	UNUSED_PARAM(command);

	::LCD_ScreenShot ();
}


/***********************************************************************
 *
 * FUNCTION:	DoInstallOther
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoInstallOther (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	static const char	filters[] =
	{
		"Palm OS Files (*.prc, *.pdb, *.pqa)\0" "*.prc;*.pdb;*.pqa\0"
		"Palm OS Applications (*.prc)\0"		"*.prc\0"
		"Palm OS Databases (*.pdb)\0"			"*.pdb\0"
		"Palm OS Clippings (*.pqa)\0"			"*.pqa\0"
		"All Files (*.*)\0" 					"*.*\0"
		""
	};

	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	EmFileRef	fileRef = gEmuPrefs->GetIndPRCMRU (0);

	char	drive [_MAX_DRIVE];
	char	dir [_MAX_DIR];
	char	fName [_MAX_FNAME];
	char	ext [_MAX_EXT];
	char	path [_MAX_PATH];

	_splitpath (fileRef.GetFullPath ().c_str (), drive, dir, fName, ext);
	_makepath (path, drive, dir, NULL, NULL);

	reply.lStructSize		= sizeof (OPENFILENAME);
	reply.hwndOwner 		= NULL; 	// hwnd; -- Don't do this, or dialog will be too small!
	reply.hInstance 		= gInstance;
	reply.lpstrFilter		= filters;
	reply.nFilterIndex		= 1;
	reply.lpstrFile 		= (LPTSTR) malloc (MAX_PATH);
	reply.nMaxFile			= MAX_PATH;
	reply.lpstrInitialDir	= path;
	reply.Flags 			= OFN_ALLOWMULTISELECT | OFN_ENABLEHOOK | OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_NOCHANGEDIR;
	reply.lpstrDefExt		= "prc";
	reply.lpfnHook			= &::OpenMultiFileProc;

	reply.lpstrFile[0] = 0;

	if (::GetOpenFileName (&reply))
	{
		EmFileRefList	fileList;

		POSITION	pos = ::GetStartPosition (reply);
		string		pathName;
		while (pos)
		{
			pathName = ::GetNextPathName (reply, pos);
			fileList.push_back (EmFileRef (pathName));
		}

		::PrvLoadPalmFiles (fileList);
	}

	free (reply.lpstrFile);
}


/***********************************************************************
 *
 * FUNCTION:	DoInstallRecent
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoInstallRecent (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(hwnd);
	UNUSED_PARAM(command);

	if (command >= ID_INSTALL_RECENT &&
		command < ID_INSTALL_RECENT + MRU_COUNT)
	{
		EmFileRef	theRef = gEmuPrefs->GetIndPRCMRU (command - ID_INSTALL_RECENT);

		if (theRef.IsSpecified ())
		{
			EmFileRefList	fileList;
			fileList.push_back (theRef);
			::PrvLoadPalmFiles (fileList);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoExportDatabase
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoExportDatabase (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(hwnd);
	UNUSED_PARAM(command);

	DatabaseInfo	pdb;

	int result = ::DialogBoxParam (	gInstance,
									MAKEINTRESOURCE (IDD_EXPORTDATABASE),
									hwnd,
									&::ExportDatabaseProc,
									(LPARAM) &pdb);

	if (result == IDOK)
	{
		UInt16		attributes;
		ErrCode		err = ::DmDatabaseInfo (pdb.cardNo, pdb.dbID, NULL,
							&attributes, NULL, NULL, NULL, NULL,
							NULL, NULL, NULL, NULL, NULL);
		Errors::ThrowIfPalmError(err);

		Bool		isResourceFile = (attributes & dmHdrAttrResDB) != 0;

		OPENFILENAME	reply;
		memset (&reply, 0, sizeof(reply));

		string			prompt ("Save file as...");
		char			fileName[MAX_PATH];

		strcpy (fileName, pdb.dbName);

		if (isResourceFile)
			strcat (fileName, ".prc");
		else
			strcat (fileName, ".pdb");

		reply.lStructSize		= sizeof (OPENFILENAME);
		reply.hwndOwner 		= NULL; 	// hwnd; -- Don't do this, or dialog will be too small!
		reply.hInstance 		= gInstance;
		reply.lpstrFilter		= "Palm OS Files (*.prc, *.pdb, *.pqa)\0" "*.prc;*.pdb;*.pqa\0";
		reply.nFilterIndex		= 1;
		reply.lpstrFile 		= fileName;
		reply.nMaxFile			= MAX_PATH;
		reply.lpstrInitialDir	= NULL;
		reply.lpstrTitle		= prompt.c_str ();
		reply.Flags 			= OFN_ENABLEHOOK | OFN_OVERWRITEPROMPT | OFN_EXPLORER;
		reply.lpstrDefExt		= isResourceFile ? "prc" : "pdb";
		reply.lpfnHook			= CenterDlgProc;

		if (::GetSaveFileName (&reply))
		{
			EmFileRef		ref (fileName);
			EmStreamFile	stream (ref, kCreateOrEraseForUpdate);

			::SavePalmFile (stream, pdb.cardNo, pdb.dbName);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	DoHotSync
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoHotSync (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(hwnd);
	UNUSED_PARAM(command);

	Hardware::HotSyncEvent (true);
	gHotSyncDownStart = ::GetTickCount ();
}


/***********************************************************************
 *
 * FUNCTION:	DoReset
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoReset (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(hwnd);
	UNUSED_PARAM(command);

	Bool	dummy;
	Platform::GetCPU ()->StopThread (dummy, dummy, kStopOnADime);

	Emulator::Reset ();

	Platform::GetCPU ()->StartThread ();
}


/***********************************************************************
 *
 * FUNCTION:	DoTransferROM
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoTransferROM (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	EmROMTransfer::ROMTransfer ();
}


/***********************************************************************
 *
 * FUNCTION:	DoGremlinsNew
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoGremlinsNew (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	EmDlg::DoHordeNew ();
}


/***********************************************************************
 *
 * FUNCTION:	DoGremlinsStep
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoGremlinsStep (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	Hordes::Step ();
}


/***********************************************************************
 *
 * FUNCTION:	DoGremlinsResume
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoGremlinsResume (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	Hordes::Resume ();
}


/***********************************************************************
 *
 * FUNCTION:	DoGremlinsStop
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoGremlinsStop (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	Hordes::Stop ();
}

void DoHordesFromControlFile(HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	static const char	filters[] =
	{
		"Gremlin Hordes Control File (*.dat)\0" "*.dat\0"
		"All Files (*.*)\0" 					"*.*\0"
		""
	};

	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	EmFileRef	fileRef = gEmuPrefs->GetIndPRCMRU (0);

	char	drive [_MAX_DRIVE];
	char	dir [_MAX_DIR];
	char	fName [_MAX_FNAME];
	char	ext [_MAX_EXT];
	char	path [_MAX_PATH];

	_splitpath (fileRef.GetFullPath ().c_str (), drive, dir, fName, ext);
	_makepath (path, drive, dir, NULL, NULL);

	reply.lStructSize		= sizeof (OPENFILENAME);
	reply.hwndOwner 		= NULL; 	// hwnd; -- Don't do this, or dialog will be too small!
	reply.hInstance 		= gInstance;
	reply.lpstrFilter		= filters;
	reply.nFilterIndex		= 1;
	reply.lpstrFile 		= (LPTSTR) malloc (MAX_PATH);
	reply.nMaxFile			= MAX_PATH;
	reply.lpstrInitialDir	= path;
	reply.Flags 			= OFN_ENABLEHOOK | OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_NOCHANGEDIR;
	reply.lpstrDefExt		= "prc";
	reply.lpfnHook			= &::OpenMultiFileProc;

	reply.lpstrFile[0] = 0;

	if (::GetOpenFileName (&reply))
	{
		EmFileRef	controlFile (reply.lpstrFile);
		Hordes::SetGremlinStatePathFromControlFile (controlFile);
		Hordes::ResumeSearchProgress (controlFile);
	}

	free (reply.lpstrFile);
}



/***********************************************************************
 *
 * FUNCTION:	DoProfileStart
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoProfileStart (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(hwnd);
	UNUSED_PARAM(command);

	if (!gProfilingEnabled)
		::ProfileInit (MAXFNCALLS, MAXDEPTH);

	::ProfileStart ();
}


/***********************************************************************
 *
 * FUNCTION:	DoProfileStop
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoProfileStop (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(hwnd);
	UNUSED_PARAM(command);

	::ProfileStop ();
}


/***********************************************************************
 *
 * FUNCTION:	DoProfileDump
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoProfileDump (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(hwnd);
	UNUSED_PARAM(command);

	::ProfileDump (NULL);
	::ProfileCleanup ();
}

/***********************************************************************
 *
 * FUNCTION:	DoProperties
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoProperties (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	gHostPrefs->EditProperties (hwnd);
}


/***********************************************************************
 *
 * FUNCTION:	DoLoggingOptions
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoLoggingOptions (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	EmDlg::DoEditLoggingOptions (kNormalLogging);
}


/***********************************************************************
 *
 * FUNCTION:	DoDebugOptions
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoDebugOptions (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	gHostPrefs->EditDebugOptions (hwnd);
}


/***********************************************************************
 *
 * FUNCTION:	DoTracingOptions
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoTracingOptions (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	gHostPrefs->EditTracingOptions (hwnd);
}


/***********************************************************************
 *
 * FUNCTION:	DoSkins
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoSkins (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	EmDlg::DoEditSkins ();
}


/***********************************************************************
 *
 * FUNCTION:	DoBreakpoints
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoBreakpoints (HWND hwnd, WPARAM command)
{
	UNUSED_PARAM(command);

	::DoBreakpointsDialog (hwnd);
}


/***********************************************************************
 *
 * FUNCTION:	DoMinimize
 *
 * DESCRIPTION:	Minimize the emulator window.
 *
 * PARAMETERS:	standard Windows interface parameters
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoMinimize (HWND hwnd, WPARAM command)
{
	ShowWindow(hwnd, SW_MINIMIZE);
}


/***********************************************************************
 *
 * FUNCTION:	DoExit
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void DoExit (HWND hwnd, WPARAM command)
{
	if (::PrvCloseSession (hwnd))
	{
		Startup::ScheduleQuit();
	}
}


/***********************************************************************
 *
 * FUNCTION:	PopupMenu
 *
 * DESCRIPTION:	Shows standard menu.
 *
 * PARAMETERS:	standard Windows interface parameters.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void PrvAddRecentSessionFiles (HMENU openMenu)
{
	// Add recent .psf names.

	string	emptyString (Platform::GetString (IDS_EMPTY));

	for (int ii = 0; ; ++ii)
	{
		EmFileRef	fileRef = gEmuPrefs->GetIndRAMMRU (ii);
		if (!fileRef.IsSpecified ())
			break;

		char	buf[MAX_PATH];
		sprintf (buf, "&%d %s", ii + 1, fileRef.GetFullPath ().c_str());
		::AppendMenu (openMenu, MF_STRING, ID_OPEN_RECENT + ii, buf);
	}

	if (ii == 0)
	{
		::AppendMenu (openMenu, MF_STRING | MF_GRAYED, 0, emptyString.c_str ());
	}
}

void PrvAddRecentPalmFiles (HMENU installMenu)
{
	// Add recent .prc names.

	string	emptyString (Platform::GetString (IDS_EMPTY));

	for (int ii = 0; ; ++ii)
	{
		EmFileRef	fileRef = gEmuPrefs->GetIndPRCMRU (ii);
		if (!fileRef.IsSpecified ())
			break;

		char	buf[MAX_PATH];
		sprintf (buf, "&%d %s", ii + 1, fileRef.GetFullPath ().c_str());
		::AppendMenu (installMenu, MF_STRING, ID_INSTALL_RECENT + ii, buf);
	}

	if (ii == 0)
	{
		::AppendMenu (installMenu, MF_STRING | MF_GRAYED, 0, emptyString.c_str ());
	}
}

HMENU PrvCreatePartiallyBoundMenu (void)
{
	const int	MENU_OPEN		= 3;
	const int	MENU_INSTALL	= 9;

	HMENU	rootMenu = ::LoadMenu (gInstance, MAKEINTRESOURCE (IDR_MENU_PARTIALLY_BOUND));
	assert (rootMenu);

	HMENU	menu = ::GetSubMenu (rootMenu, 0);
	assert (menu);

	HMENU	openMenu = ::GetSubMenu (menu, MENU_OPEN);
	assert (openMenu);

	HMENU	installMenu = ::GetSubMenu (menu, MENU_INSTALL);
	assert (installMenu);

	::PrvAddRecentSessionFiles (openMenu);
	::PrvAddRecentPalmFiles (installMenu);

	return rootMenu;
}

HMENU PrvCreateFullyBoundMenu (void)
{
	const int	MENU_INSTALL	= 4;

	HMENU	rootMenu = ::LoadMenu (gInstance, MAKEINTRESOURCE (IDR_MENU_FULLY_BOUND));
	assert (rootMenu);

	HMENU	menu = ::GetSubMenu (rootMenu, 0);
	assert (menu);

	HMENU	installMenu = ::GetSubMenu (menu, MENU_INSTALL);
	assert (installMenu);

	::PrvAddRecentPalmFiles (installMenu);

	return rootMenu;
}

HMENU PrvCreateMenu (void)
{
	const int	MENU_OPEN		= 3;
	const int	MENU_INSTALL	= 11;
	const int	MENU_GREMLINS	= 17;
	const int	MENU_PROFILE	= 18;
	const int	MENU_SETTINGS	= 19;

	HMENU	rootMenu = ::LoadMenu (gInstance, MAKEINTRESOURCE (IDR_MENU));
	assert (rootMenu);

	HMENU	menu = ::GetSubMenu (rootMenu, 0);
	assert (menu);

	HMENU	openMenu = ::GetSubMenu (menu, MENU_OPEN);
	assert (openMenu);

	HMENU	installMenu = ::GetSubMenu (menu, MENU_INSTALL);
	assert (installMenu);

	HMENU	gremlinsMenu = ::GetSubMenu (menu, MENU_GREMLINS);
	assert (gremlinsMenu);

	HMENU	profileMenu = ::GetSubMenu (menu, MENU_PROFILE);
	assert (profileMenu);

	HMENU	settingsMenu = ::GetSubMenu (menu, MENU_SETTINGS);
	assert (settingsMenu);

	::PrvAddRecentSessionFiles (openMenu);
	::PrvAddRecentPalmFiles (installMenu);

#if !HAS_PROFILING
	// Enable/disable menu items.

	::EnableMenuItem (profileMenu, ID_PROFILE_START, MF_GRAYED);
	::EnableMenuItem (profileMenu, ID_PROFILE_STOP, MF_GRAYED);
	::EnableMenuItem (profileMenu, ID_PROFILE_DUMP, MF_GRAYED);
#endif

	return rootMenu;
}


void PopupMenu (HWND hwnd, int x, int y)
{
	HMENU	rootMenu;

	if (::IsBoundPartially ())
	{
		rootMenu = ::PrvCreatePartiallyBoundMenu ();
	}
	else if (::IsBoundFully ())
	{
		rootMenu = ::PrvCreateFullyBoundMenu ();
	}
	else
	{
		rootMenu = ::PrvCreateMenu ();
	}
	// Show the menu.

	POINT	p = { x,  y };
	::ClientToScreen (hwnd, &p);

	HMENU	menu = ::GetSubMenu (rootMenu, 0);
	::TrackPopupMenu (menu, TPM_LEFTALIGN | TPM_RIGHTBUTTON, p.x, p.y, 0, hwnd, NULL);

	::DestroyMenu (rootMenu);
}


/***********************************************************************
 *
 * FUNCTION:    FUNCTION_NAME
 *
 * DESCRIPTION: DESCRIPTION
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

static void PrvOnChar (HWND hwnd, TCHAR ch, int cRepeat)
{
	switch (ch)
	{
		case VK_RETURN: ch = chrLineFeed;		break;

			// These are handled in the key up/down handlers
			// below, so they shouldn't result in a key being
			// inserted into the event queue.
			//
			// Actually...pressing most of these keys doesn't
			// result in a WM_CHAR event.  We only need to check
			// for VK_ESCAPE.
//		case VK_UP: 	return;
//		case VK_DOWN:	return;
//		case VK_LEFT:	return;
//		case VK_RIGHT:	return;
		case VK_ESCAPE: return;
//		case VK_PRIOR:	return;
//		case VK_NEXT:	return;
//		case VK_F1: 	return;
//		case VK_F2: 	return;
//		case VK_F3: 	return;
//		case VK_F4: 	return;
	}

	Hardware::KeyboardEvent (ch);
}


static void PrvOnClose (HWND hwnd)
{
	::DoExit (hwnd, 0);
}


static void PrvOnCommand (HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	int index = 0;
	while (kCommandHandlers[index].fCommand)
	{
		if (kCommandHandlers[index].fCommand == id)
		{
			try
			{
				if (kCommandHandlers[index].fStopType)
				{
					CPUStopper	stopper (kCommandHandlers[index].fStopType);
					if (stopper.Stopped ())
					{
						kCommandHandlers[index].fHandler (hwnd, id);
					}
					else
					{
						stopper.ShowError ("perform the command");
					}
				}
				else
				{
					kCommandHandlers[index].fHandler (hwnd, id);
				}
			}
			catch (ErrCode errCode)
			{
				Errors::ReportIfError (kCommandHandlers[index].fOperation, errCode, 0, false);
			}

			break;
		}

		index++;
	}
}


static BOOL PrvOnCreate (HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
	::LCD_ResetWindow (hwnd);
	::DragAcceptFiles (hwnd, TRUE);
	return true;
}


static void PrvOnDestroy (HWND hwnd)
{
	::LCD_SetWindow (NULL);
	Bool dummy;
	Platform::GetCPU ()->StopThread (dummy, dummy, kStopOnADime);
	::DragAcceptFiles (hwnd, FALSE);

	{
		RECT	wr;
		::GetWindowRect (hwnd, &wr);

		PointType	pt = { wr.left, wr.top };
		Preference <PointType>	pref (kPrefKeyWindowLocation);
		pref = pt;
	}

	// Windows has already destroyed all child windows at this point and the 
	// GCW's WM_DESTROY handler sets the global handle to the window to NULL,
	// so we're already set. This call wouldn't work anyway since it posts a
	// message that won't be answered because the GCW is already gone.

	// Platform::GCW_Close ();

	::KillTimer (hwnd, 1);
	::PostQuitMessage (0);
}


static void PrvOnDisplayChange (HWND hwnd, UINT bitsPerPixel, UINT cxScreen, UINT cyScreen)
{
	::LCD_DisplayChanged (hwnd);
}


static void PrvOnDropFiles (HWND hwnd, HDROP hdrop)
{
	// A file is being dropped.

	HDROP	hDropInfo = hdrop;

	// Get the number of files.

	int numFiles = ::DragQueryFile (hDropInfo, (DWORD)(-1), NULL, 0);

	// This would get the drop point, if we needed it.

//	POINT	pt;
//	::DragQueryPoint (hDropInfo, &pt);

	EmFileRefList	fileList;

	for (int ii = 0; ii < numFiles; ++ii)
	{
		char	filePath [MAX_PATH];
		::DragQueryFile (hDropInfo, ii, filePath, sizeof (filePath));

		EmFileRef	fileRef (filePath);
		fileList.push_back (fileRef);
	}

	// Signal that the drag-and-drop operation is over.

	::DragFinish (hDropInfo);

	LoadAnyFiles (fileList);
}


static void PrvOnKeyDown (HWND hwnd, UINT vk, BOOL fDown, int cRepeat, UINT flags)
{
	unsigned char		whichKey	= 0;
	SkinElementType 	whichButton = kElement_None;

	switch (vk)
	{
		case VK_UP: 	whichKey = chrUpArrow;		break;
		case VK_DOWN:	whichKey = chrDownArrow;	break;
		case VK_LEFT:	whichKey = chrLeftArrow;	break;
		case VK_RIGHT:	whichKey = chrRightArrow;	break;

		case VK_ESCAPE: whichButton = kElement_PowerButton; 	break;
		case VK_PRIOR:	whichButton = kElement_UpButton;		break;
		case VK_NEXT:	whichButton = kElement_DownButton;		break;
		case VK_F1: 	whichButton = kElement_App1Button;		break;
		case VK_F2: 	whichButton = kElement_App2Button;		break;
		case VK_F3: 	whichButton = kElement_App3Button;		break;
		case VK_F4: 	whichButton = kElement_App4Button;		break;
	}

	if (whichKey != 0)
	{
		Hardware::KeyboardEvent (whichKey);
	}

	if (whichButton != kElement_None)
	{
		Hardware::ButtonEvent (true, whichButton);
	}
}


static void PrvOnKeyUp (HWND hwnd, UINT vk, BOOL fDown, int cRepeat, UINT flags)
{
	SkinElementType 	whichButton = kElement_None;

	switch (vk)
	{
		case VK_ESCAPE: whichButton = kElement_PowerButton; 	break;
		case VK_PRIOR:	whichButton = kElement_UpButton;		break;
		case VK_NEXT:	whichButton = kElement_DownButton;		break;
		case VK_F1: 	whichButton = kElement_App1Button;		break;
		case VK_F2: 	whichButton = kElement_App2Button;		break;
		case VK_F3: 	whichButton = kElement_App3Button;		break;
		case VK_F4: 	whichButton = kElement_App4Button;		break;
	}

	if (whichButton != kElement_None)
	{
		Hardware::ButtonEvent (false, whichButton);
	}
}


static void PrvOnLButtonDown (HWND hwnd, BOOL fDoubleClick, int x, int y, UINT keyFlags)
{
	::LCD_PenEvent (hwnd, x, y, true);
}


static void PrvOnLButtonUp (HWND hwnd, int x, int y, UINT keyFlags)
{
	::LCD_PenEvent (hwnd, x, y, false);
}


static void PrvOnMouseMove (HWND hwnd, int x, int y, UINT keyFlags)
{
	if (GetCapture ())
	{
		::LCD_PenEvent (hwnd, x, y, true);
	}
}


static UINT PrvOnNcHitTest (HWND hwnd, int x, int y)
{
	POINT	clientPt = { x, y };
	::ScreenToClient (hwnd, &clientPt);

	PointType		pt = { clientPt.x, clientPt.y };
	SkinElementType what = ::SkinTestPoint (pt);

	if (what == kElement_Frame)
		return HTCAPTION;

	return FORWARD_WM_NCHITTEST (hwnd, x, y, DefWindowProc);
}


static void PrvOnNcRButtonDown (HWND hwnd, BOOL fDoubleClick, int x, int y, UINT codeHitTest)
{
	// When the user clicks with the right mouse button, popup our menu.

	POINT	pt = { x, y };
	::ScreenToClient (hwnd, &pt);
	::PopupMenu (hwnd, pt.x, pt.y);
}


static void PrvOnPaint (HWND hwnd)
{
	::LCD_Paint (hwnd);
}


static void PrvOnPaletteChanged(HWND hwnd, HWND hwndPaletteChange)
{
	if (hwndPaletteChange != hwnd)
	{
		(void) ::LCD_SelectPalette (hwnd);
	}
}


static BOOL PrvOnQueryNewPalette(HWND hwnd)
{
	int n = ::LCD_SelectPalette (hwnd);

	if (n > 0)
	{
		// Invalidate window
	}

	// If the window realizes its logical palette, it must return TRUE;
	// otherwise, it must return FALSE.

	return n != -1;
}


static void PrvOnRButtonDown (HWND hwnd, BOOL fDoubleClick, int x, int y, UINT keyFlags)
{
	// When the user clicks with the right mouse button, popup our menu.

	POINT	pt = { x, y };
	::PopupMenu (hwnd, pt.x, pt.y);
}


static void PrvOnSysKeyDown (HWND hwnd, UINT vk, BOOL fDown, int cRepeat, UINT flags)
{
	switch (vk)
	{
		// When the user presses F10, popup our menu.

		case VK_F10:
			::PopupMenu (hwnd, 0, 0);
			return;

		// When the user presses Alt+F11, minimize our window.

		case VK_F11:
			if (flags & (1 << (29 - 16)))
				::DoMinimize (hwnd, 0);
			return;
	}

	FORWARD_WM_SYSKEYDOWN (hwnd, vk, cRepeat, flags, DefWindowProc);
}


static void PrvOnTimer (HWND hwnd, UINT id)
{
	// Perform periodic tasks.

	// Update our window with the current contents of the emulated LCD buffer.

	{
		CPUStopper	stopper (kStopOnADime);
		if (stopper.Stopped ())
		{
			::LCD_Refresh (hwnd);
		}
	}

	// Idle our sockets.

	CSocket::IdleAll ();

	// Idle any packets that are looking for signals.
	// That is, see if any have timed out by now.

	RPC::Idle ();

	// If we need to start a new Gremlin Horde, and the OS
	// is now booted enough to handle our mucking about as
	// we start it up, then start starting it up.

	if (Patches::EvtGetEventCalled ())
	{
		if (Startup::NewHorde (NULL))
		{
			CPUStopper	stopper (kStopOnATrap);
			if (stopper.Stopped ())
			{
				HordeInfo	info;
				Startup::NewHorde (&info);
				Hordes::New (info);
			}
		}
	}

	if (gOpenGCWWindow)
	{
		gOpenGCWWindow = false;
		Platform::GCW_Open ();
	}


	// If we're emulating the user pressing the hotsync button, emulate
	// releasing the button after 1 second.

	if (gHotSyncDownStart && ::GetTickCount () - gHotSyncDownStart > 1000)
	{
		CPUStopper	stopper (kStopOnADime);
		if (stopper.Stopped ())
		{
			Hardware::HotSyncEvent (false);
			gHotSyncDownStart = 0;
		}
	}

	// Bump this guy so that periodic tasks can be handled in
	// Hardware::Cycle.

	gPseudoTickcount++;

	if (Patches::GetQuitStage () == kTimeToQuit)
	{
		// Reset fgQuitStage so that if DoExit brings up a "Save this?"
		// dialog, we don't re-enter this code at the dialogs idle time.

		Patches::SetQuitStage (kNoQuit);
		::DoExit (hwnd, ID_EXIT);
	}

	EmFileRef	f;
	if (Startup::CloseSession (f))
	{
		CloseActionType	action = f.IsSpecified() ? kSaveAlways : kSaveNever;
		if (::PrvCloseSession (hwnd, f, action))
		{
			// Don't do this, in case the request to close was followed
			// by a request to quit.
//			Startup::ScheduleAskWhatToDo();
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	DisplayWndProc
 *
 * DESCRIPTION:	Dispatch the command to the right handler via the
 *				message cracker macros.
 *
 * PARAMETERS:	Standard Window DefProc parameters.
 *
 * RETURNED:	Result code specific to the command.
 *
 ***********************************************************************/

#define MY_HANDLE_MSG(hwnd, message, fn, how)							\
    case (message):														\
		if (how)														\
		{																\
			CPUStopper	stopper(how);									\
			return HANDLE_##message((hwnd), (wParam), (lParam), (fn));	\
		}																\
		else															\
			return HANDLE_##message((hwnd), (wParam), (lParam), (fn))

LRESULT CALLBACK DisplayWndProc (HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
		MY_HANDLE_MSG (hWnd, WM_CHAR,				PrvOnChar,				kStopOnADime);
		MY_HANDLE_MSG (hWnd, WM_CLOSE,				PrvOnClose,				0);
		MY_HANDLE_MSG (hWnd, WM_COMMAND,			PrvOnCommand,			0);
		MY_HANDLE_MSG (hWnd, WM_CREATE,				PrvOnCreate,			0);
		MY_HANDLE_MSG (hWnd, WM_DESTROY,			PrvOnDestroy,			0);
		MY_HANDLE_MSG (hWnd, WM_DISPLAYCHANGE,		PrvOnDisplayChange,		0);
		MY_HANDLE_MSG (hWnd, WM_DROPFILES,			PrvOnDropFiles,			kStopOnATrap);
		MY_HANDLE_MSG (hWnd, WM_KEYDOWN,			PrvOnKeyDown,			kStopOnADime);
		MY_HANDLE_MSG (hWnd, WM_KEYUP,				PrvOnKeyUp,				kStopOnADime);
		MY_HANDLE_MSG (hWnd, WM_LBUTTONDOWN,		PrvOnLButtonDown,		kStopOnADime);
		MY_HANDLE_MSG (hWnd, WM_LBUTTONUP,			PrvOnLButtonUp,			kStopOnADime);
		MY_HANDLE_MSG (hWnd, WM_MOUSEMOVE,			PrvOnMouseMove,			kStopOnADime);
		MY_HANDLE_MSG (hWnd, WM_NCHITTEST,			PrvOnNcHitTest,			0);
		MY_HANDLE_MSG (hWnd, WM_NCRBUTTONDOWN,		PrvOnNcRButtonDown,		0);
		MY_HANDLE_MSG (hWnd, WM_PAINT,				PrvOnPaint,				kStopOnADime);
		MY_HANDLE_MSG (hWnd, WM_PALETTECHANGED,		PrvOnPaletteChanged,	kStopOnADime);
		MY_HANDLE_MSG (hWnd, WM_QUERYNEWPALETTE,	PrvOnQueryNewPalette,	kStopOnADime);
		MY_HANDLE_MSG (hWnd, WM_RBUTTONDOWN,		PrvOnRButtonDown,		0);
		MY_HANDLE_MSG (hWnd, WM_SYSKEYDOWN,			PrvOnSysKeyDown,		0);
		MY_HANDLE_MSG (hWnd, WM_TIMER,				PrvOnTimer,				0);

		case MM_WOM_DONE:
			gSounds->NextSound();
			return NULL;
	}

	return ::DefWindowProc (hWnd, msg, wParam, lParam);
}


#if (TIE_TICKS_TO_HOST)
Bool	gKillTimerThread;

VOID CALLBACK TimerProc (
	HWND hwnd,		// handle of window for timer messages
	UINT uMsg,		// WM_TIMER message
	UINT idEvent,	// timer identifier
	DWORD dwTime)	// current system time)
{
	UNUSED_PARAM(hwnd);
	UNUSED_PARAM(uMsg);
	UNUSED_PARAM(idEvent);

	HWRegisters::SetCurrentMilliseconds (dwTime /* ::GetTickCount ()*/);

	if (gKillTimerThread)
	{
		::PostQuitMessage (0);
	}
}


DWORD CALLBACK TimerThread (void*)
{
	UINT	timerID = ::SetTimer (NULL, 0, 10, TimerProc);

	MSG msg;
	while (::GetMessage (&msg, NULL, 0, 0))
	{
		::TranslateMessage (&msg);
		::DispatchMessage (&msg);
	}

	::KillTimer (NULL, timerID);

	return 0;
}
#endif


/***********************************************************************
 *
 * FUNCTION:	DoUIThread
 *
 * DESCRIPTION:	Thread proc for the UI thread.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

DWORD CALLBACK DoUIThread (void*)
{
	// Open the connection to the external debugger.  We used to do
	// this from Platform::ConnectToDebugger, but that would get
	// called from a different thread.  Because the window used for
	// communications would belong to a different thread, the message
	// loop below would not process messages for it.  So we do it here.

	::MMFOpenFiles (gInstance, CMMFSocket::HandlePacketCB, 0,
					kFromEmulatorFileName, kToEmulatorFileName);

	// Kickstart the CPU emulator (running in its own thread).  Do this
	// *before* creating the LCD window, in case any messages are sent
	// to the window that require the use of the CPU thread.

	Platform::GetCPU ()->StartThread ();

	// Create the LCD window.

	::LCD_Init ();

	// Create SoundMaker object.

	gSounds = new WinSounds;

#if (TIE_TICKS_TO_HOST)
	// Start up a little timer thread used to keep the emulated tickcount
	// in sync with the host's

	gKillTimerThread = false;
	DWORD	tid;
	HANDLE	timer = ::CreateThread (	NULL,			// Security attributes
										0,				// Stack size
										TimerThread,	// Start address
										NULL,			// Thread-func parameter
										0,				// Creation flags
										&tid);			// Thread ID
#endif

	// Our main event loop.  Retrieve and handle events until the application quits.

	MSG msg;
	while (::GetMessage (&msg, NULL, 0, 0))
	{
		::TranslateMessage (&msg);
		::DispatchMessage (&msg);
	}

#if (TIE_TICKS_TO_HOST)
	// Kill the timer thread.

	gKillTimerThread = true;
	::WaitForSingleObject (timer, INFINITE);	// Wait for task to die.
	::CloseHandle (timer);
#endif

	::MMFCloseFiles (gInstance);

	delete gSounds;
	gSounds = NULL;

	::LCD_Quit ();

	// Exit the thread.

	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	StartUIThread
 *
 * DESCRIPTION:	Create and return the handle to the UI thread.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

HANDLE StartUIThread (void* data)
{
	DWORD	tid;
	HANDLE	thread = ::CreateThread (	NULL,			// Security attributes
										0,				// Stack size
										DoUIThread, 	// Start address
										data,			// Thread-func parameter
										0,				// Creation flags
										&tid);			// Thread ID
	return thread;
}
