/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Document.h"

#include "Application.h"		// gOpenGCWWindow
#include "CPU_MT.h"				// CPU_init, CPU_load
#include "CPU_REG.h"			// Emulator::Save
#include "EmDlgWin.h"			// CenterWindow, CenterDlgProc
#include "Emulator.h"			// gInstance
#include "Miscellaneous.h"		// kDevicePilot1000, etc.
#include "Platform.h"			// Platform::ReportError
#include "PreferenceMgr.h"		// Preference
#include "Strings.r.h"			// kStr_CmdSave values
#include "Hordes.h"				// Hordes::PostLoad()

#include "resource.h"


// ======================================================================
//	Globals and constants
// ======================================================================

static Configuration	gCfg;
static DeviceType		gDevice;
static SkinNameList		gSkinChoices;


// ======================================================================
//	Private functions
// ======================================================================

static BOOL CALLBACK	PrvQueryNewDocumentDlgProc (HWND hwnd,
													UINT msg,
													WPARAM wparam,
													LPARAM lparam);
static Bool		PrvBrowseForROM			(HWND hwnd, EmFileRef& romFile);

static void		PrvDrawPopupButton		(LPDRAWITEMSTRUCT lpDIS);
static void		PrvClickPopupButton		(LPDRAWITEMSTRUCT lpDIS, HWND dlg);

static void		PrvClickDevicePopup		(LPDRAWITEMSTRUCT lpDIS, HWND dlg);
static void		PrvClickSkinPopup		(LPDRAWITEMSTRUCT lpDIS, HWND dlg);
static void		PrvClickRAMSizePopup	(LPDRAWITEMSTRUCT lpDIS, HWND dlg);
static UINT		PrvClickCommon			(LPDRAWITEMSTRUCT lpDIS, HWND dlg, HMENU menu);

static void		PrvCheckSkinPopup		(HWND hwnd);

static void		PrvSetDevicePopupText	(HWND hwnd, DeviceType);
static void		PrvSetSkinPopupText		(HWND hwnd, const string&);
static void		PrvSetRAMSizePopupText	(HWND hwnd, RAMSizeType);
static void		PrvSetROMFileButtonText	(HWND hwnd, const EmFileRef& romFile);

static string	PrvGetDeviceText		(DeviceType);
static string	PrvGetRAMSizeText		(RAMSizeType ramSize);


/***********************************************************************
 *
 * FUNCTION:	QueryNewDocument
 *
 * DESCRIPTION:	Ask the user for the parameters for a new session.
 *				This information includes: devide type, country
 *				specific graphics, ram size, and rom file.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

Bool QueryNewDocument (HWND hwnd, Configuration& cfg)
{
	// If we have an embedded ROM, get the configuration settings
	// from the embedded configuration.

	if (::IsBound ())
	{
		cfg.fDeviceType	= Platform::GetBoundDeviceType ();
		cfg.fRAMSize	= Platform::GetBoundRAMSize ();

		return true;
	}

	assert (!::IsBound ());

	gCfg = cfg;

	int result = ::DialogBox (gInstance, MAKEINTRESOURCE (IDD_NEWFILE), hwnd,
								&::PrvQueryNewDocumentDlgProc); 

	if (result == IDOK)
	{
		cfg = gCfg;

		Preference<SkinNameList>	prefSkins (kPrefKeySkins);
		prefSkins = gSkinChoices;
	}

	return result == IDOK;
}


/***********************************************************************
 *
 * FUNCTION:	PrvQueryNewDocumentDlgProc
 *
 * DESCRIPTION:	DlgProc for QueryNewDocument
 *
 * PARAMETERS:	Standard DlgProc parameters.
 *
 *				Custom information is communications in the global
 *				gCfg variable.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

BOOL CALLBACK PrvQueryNewDocumentDlgProc (	HWND hwnd,
											UINT msg,
											WPARAM wparam,
											LPARAM lparam)
{
	UNUSED_PARAM(lparam);

	switch (msg)
	{
		case WM_INITDIALOG:
		{
			::CenterWindow (hwnd);

			Preference<SkinNameList>	prefSkins (kPrefKeySkins);
			gSkinChoices = *prefSkins;

			::PrvSetDevicePopupText (hwnd, gCfg.fDeviceType);
			::PrvSetRAMSizePopupText (hwnd, gCfg.fRAMSize);
			::PrvSetROMFileButtonText (hwnd, gCfg.fROMFile);

			::PrvCheckSkinPopup (hwnd);

			break;
		}

		case WM_DRAWITEM:
		{
			LPDRAWITEMSTRUCT	lpDIS = (LPDRAWITEMSTRUCT) lparam;

			if (lpDIS->CtlType == ODT_BUTTON)
			{
				if (lpDIS->itemAction == ODA_DRAWENTIRE)
				{
					::PrvDrawPopupButton (lpDIS);
					return TRUE;
				}
				else if (lpDIS->itemAction == ODA_SELECT)
				{
					if ((lpDIS->itemState & ODS_SELECTED) != 0)
					{
						::PrvClickPopupButton (lpDIS, hwnd);
						return FALSE;	// TRUE;
					}
				}
			}

			break;
		}

		case WM_COMMAND:
		{
			switch (GET_WM_COMMAND_ID (wparam, lparam))
			{
				case IDOK:
				case IDCANCEL:
					::EndDialog (hwnd, LOWORD (wparam));
					break;

				case IDC_BROWSE_ROMFILE:
					if (GET_WM_COMMAND_CMD (wparam, lparam) == BN_CLICKED)
					{
						char	fileName[MAX_PATH];
						HWND	dlgItem = ::GetDlgItem (hwnd, IDC_ROMFILE);
						::GetWindowText (dlgItem, fileName, MAX_PATH);

						EmFileRef	romFile (fileName);

						if (::PrvBrowseForROM (hwnd, romFile))
						{
							HWND	dlgItem = ::GetDlgItem (hwnd, IDC_ROMFILE);
							::SetWindowText (dlgItem, romFile.GetName ().c_str ());

							gCfg.fROMFile = romFile;
						}
					}

					break;
			}

			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}


/***********************************************************************
 *
 * FUNCTION:	CreateNewDocument
 *
 * DESCRIPTION:	Create a new session based on the given configuration.
 *
 * PARAMETERS:	cfg - configuration containing the specifics of
 *					the new session.
 *
 * RETURNED:	Error code for any errors that may occur.
 *
 ***********************************************************************/

ErrCode CreateNewDocument (const Configuration& cfg)
{
	// Try initializing the emulator with all this information.

//	assert (!Platform::GetCPU ().Running ());

	ErrCode result = errNone;

	try
	{
		Emulator::Initialize (cfg);
	}
	catch (ErrCode code)
	{
		result = code;
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	QueryOpenDocument
 *
 * DESCRIPTION:	Ask the user for an existing document to open.
 *
 * PARAMETERS:	hwnd - parent window.
 *
 *				fileRef - object to receive the user's choice.
 *
 * RETURNED:	True if the user made a choice, false otherwise.
 *
 ***********************************************************************/

Bool QueryOpenDocument (HWND hwnd, EmFileRef& fileRef)
{
	assert (!::IsBoundFully ());

	// Ask the user to choose a .PSF file.

	static const char	filters[] =
	{
		"Emulator Session Files (*.psf)\0"		"*.psf\0"
		"All Files (*.*)\0" 					"*.*\0"
		""
	};

	char	fileName[MAX_PATH];
	fileName[0] = 0;

	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	char			pathName[MAX_PATH];
	char			drive[_MAX_DRIVE];
	char			dir[_MAX_DIR];

	Preference<EmFileRef>	pref (kPrefKeyLastPSF);
	EmFileRef				lastRAM = *pref;

	strcpy (pathName, lastRAM.GetFullPath ().c_str ());
	_splitpath (pathName, drive, dir, NULL, NULL);
	_makepath (pathName, drive, dir, NULL, NULL);

	reply.lStructSize		= sizeof (OPENFILENAME);
	reply.hwndOwner 		= NULL; 	// hwnd; -- Don't do this, or dialog will be too small!
	reply.hInstance 		= gInstance;
	reply.lpstrFilter		= filters;
	reply.nFilterIndex		= 1;
	reply.lpstrFile 		= fileName;
	reply.nMaxFile			= MAX_PATH;
	reply.lpstrInitialDir	= pathName;
	reply.Flags 			= OFN_ENABLEHOOK /*| OFN_ENABLESIZING*/ | OFN_EXPLORER;
	reply.lpfnHook			= &::CenterDlgProc;

	BOOL	result = ::GetOpenFileName (&reply);

	if (result)
	{
		fileRef = EmFileRef (fileName);
	}

	return result != 0;
}


/***********************************************************************
 *
 * FUNCTION:	OpenOldDocument
 *
 * DESCRIPTION:	Create a new session based on an existing .psf file.  If
 *				a file cannot be found at the given location, look in
 *				the emulator's directory for it.
 *
 * PARAMETERS:	ramFileRef - reference to the .psf file.
 *
 * RETURNED:	Error code for any errors that may occur.
 *
 ***********************************************************************/

ErrCode OpenOldDocument (const EmFileRef& ramFileRef)
{
//	assert (!Platform::GetCPU ().Running ());

	EmFileRef	refCopy (ramFileRef);
	ErrCode		result = errNone;

	if (!refCopy.Exists ())
	{
		string		fileName = refCopy.GetName ();
		EmDirRef	poserDir = EmDirRef::GetEmulatorDirectory ();
		refCopy = EmFileRef (poserDir, fileName);
	}

	try
	{
		Emulator::Load (refCopy, true, true, true);
		Hordes::PostLoad();

		// If we loaded a session file with a Gremlin running, turn
		// it off, but let the user turn it back on.

		if (Hordes::IsOn ())
		{
//			Hordes::Stop ();
			gOpenGCWWindow = true;
		}
	}
	catch (ErrCode code)
	{
		result = code;
	}

	return result;
}

/***********************************************************************
 *
 * FUNCTION:	OpenResourceDocument
 *
 * DESCRIPTION:	Opens a PSF file from the embedded resource.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	Error code for any errors that may occur.
 *
 ***********************************************************************/

ErrCode OpenResourceDocument(void)
{
	Chunk	psf;
	ErrCode result = errNone;

	assert (::IsBoundFully ());

	Platform::GetPSFResource (psf);

	try
	{
		Emulator::Load (psf.GetPointer (), psf.GetLength (), true);
		Hordes::PostLoad();
	}
	catch (ErrCode code)
	{
		result = code;
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	SaveRAMImage
 *
 * DESCRIPTION:	Save the current session to a .psf file.
 *
 * PARAMETERS:	hwnd - parent window (used when asking the user for
 *					a destination file name).
 *
 *				ramFileRef - reference to a currently existing file.
 *					Optional; may be unspecified if the current
 *					session has never been saved before.
 *
 *				askUser - True if this function should always ask the
 *					user for a destination filename.  Otherwise, the
 *					user is asked only if ramFileRef is unspecified.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

BOOL SaveRAMImage (HWND hwnd, const EmFileRef& ramFileRef, Bool askUser, Bool forAutoSave)
{
	EmFileRef	realFileRef (ramFileRef);

	if (askUser || !realFileRef.IsSpecified ())
	{
		OPENFILENAME	reply;
		memset (&reply, 0, sizeof(reply));

		char			fileName[MAX_PATH] = {0};

		string			saveRAMPrompt (Platform::GetString (IDS_SAVE_RAM_PROMPT));

		char			pathName[MAX_PATH];
		char			drive[_MAX_DRIVE];
		char			dir[_MAX_DIR];

		Preference<EmFileRef>	pref (kPrefKeyLastPSF);
		EmFileRef				lastRAM = *pref;

		strcpy (pathName, lastRAM.GetFullPath ().c_str ());
		_splitpath (pathName, drive, dir, NULL, NULL);
		_makepath (pathName, drive, dir, NULL, NULL);

		reply.lStructSize		= sizeof (OPENFILENAME);
		reply.hwndOwner 		= NULL; 	// hwnd; -- Don't do this, or dialog will be too small!
		reply.hInstance 		= gInstance;
		reply.lpstrFilter		= "Emulator Session Files (*.psf)\0*.psf\0";
		reply.nFilterIndex		= 1;
		reply.lpstrFile 		= fileName;
		reply.nMaxFile			= MAX_PATH;
		reply.lpstrInitialDir	= pathName;
		reply.lpstrTitle		= saveRAMPrompt.c_str ();
		reply.Flags 			= OFN_ENABLEHOOK | OFN_OVERWRITEPROMPT | OFN_EXPLORER;
		reply.lpstrDefExt		= "psf";
		reply.lpfnHook			= CenterDlgProc;

		if (!::GetSaveFileName (&reply))
		{
			return false;
		}

		realFileRef = EmFileRef (fileName);
	}

	try
	{
		Emulator::Save (realFileRef, !forAutoSave, !forAutoSave, !forAutoSave);
	}
	catch (ErrCode errCode)
	{
		Errors::SetParameter ("%filename", realFileRef.GetName ());
		Errors::ReportIfError (kStr_CmdSave, errCode);
		return false;
	}

	return true;
}


/***********************************************************************
 *
 * FUNCTION:	PrvBrowseForROM
 *
 * DESCRIPTION:	Ask the user for a ROM image file.
 *
 * PARAMETERS:	hwnd - parent window
 *
 *				romFile - receives the user's choice.
 *
 * RETURNED:	True if the user chose a ROM file, false otherwise.
 *
 ***********************************************************************/

Bool PrvBrowseForROM (HWND hwnd, EmFileRef& romFile)
{
	// Look for files ending in ".rom" or " rom", or starting with "rom.".

	static const char	filters[] =
	{
		"Palm OS ROM Files (*.rom)\0"		"*.rom;* rom;rom.*\0"
		"All Files (*.*)\0" 				"*.*\0"
		""
	};

	char	fileName[MAX_PATH];
	fileName[0] = 0;

	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	char	pathName[MAX_PATH];
	char	drive[_MAX_DRIVE];
	char	dir[_MAX_DIR];

	strcpy (pathName, romFile.GetFullPath ().c_str ());
	_splitpath (pathName, drive, dir, NULL, NULL);
	_makepath (pathName, drive, dir, NULL, NULL);

	reply.lStructSize		= sizeof (OPENFILENAME);
	reply.hwndOwner 		= NULL; 	// hwnd; -- Don't do this, or dialog will be too small!
	reply.hInstance 		= gInstance;
	reply.lpstrFilter		= filters;
	reply.nFilterIndex		= 1;
	reply.lpstrFile 		= fileName;
	reply.nMaxFile			= MAX_PATH;
	reply.lpstrInitialDir	= pathName;
	reply.Flags 			= OFN_ENABLEHOOK /*| OFN_ENABLESIZING*/ | OFN_EXPLORER;
	reply.lpfnHook			= CenterDlgProc;

	BOOL	result = ::GetOpenFileName (&reply);

	if (result)
	{
		romFile = EmFileRef (fileName);
	}

	return result != 0;
}


void PrvDrawPopupButton (LPDRAWITEMSTRUCT lpDIS)
{
	UINT	uStyle = DFCS_BUTTONPUSH;

	// If drawing selected, add the pushed style to DrawFrameControl.

	if (lpDIS->itemState & ODS_SELECTED)
		uStyle |= DFCS_PUSHED;

	// Draw the button frame.

	::DrawFrameControl (lpDIS->hDC, &lpDIS->rcItem, DFC_BUTTON, uStyle);
	
	// Get the button's text.

	char	strText[256];
	::GetWindowText (lpDIS->hwndItem, strText, 256);

	// Draw the button text.

	RECT	bounds = lpDIS->rcItem;
	if (lpDIS->itemState & ODS_SELECTED)
	{
		bounds.left += 1;
		bounds.top += 2;
	}

	bounds.left += 10;
	::DrawText (lpDIS->hDC, strText, strlen (strText), &bounds, DT_SINGLELINE | DT_VCENTER);

	// Draw the triangle.

	HBRUSH	brush = (HBRUSH) ::GetStockObject (BLACK_BRUSH);

	bounds = lpDIS->rcItem;
	if (lpDIS->itemState & ODS_SELECTED)
	{
		bounds.right += 1;
		bounds.top += 2;
	}

	bounds.right -= 10;
	bounds.left = bounds.right - 1;
	bounds.top = (bounds.top + bounds.bottom) / 2 + 1;
	bounds.bottom = bounds.top + 1;

	for (int yy = 0; yy < 4; ++yy)
	{
		::FillRect (lpDIS->hDC, &bounds, brush);
		bounds.left--;
		bounds.right++;
		bounds.top--;
		bounds.bottom--;
	}
}


void PrvClickPopupButton (LPDRAWITEMSTRUCT lpDIS, HWND dlg)
{
	// Prepare the menus for the various buttons, display them,
	// and handle the user's selection.

	if (lpDIS->CtlID == IDC_DEVICE)
	{
		::PrvClickDevicePopup (lpDIS, dlg);
	}
	else if (lpDIS->CtlID == IDC_SKIN)
	{
		::PrvClickSkinPopup (lpDIS, dlg);
	}
	else if (lpDIS->CtlID == IDC_RAMSIZE)
	{
		::PrvClickRAMSizePopup (lpDIS, dlg);
	}
}


void PrvClickDevicePopup (LPDRAWITEMSTRUCT lpDIS, HWND dlg)
{
	HMENU	menu = ::CreatePopupMenu ();

	EmDeviceMenuItemList			devices = EmDevice::GetMenuItemList ();
	EmDeviceMenuItemList::iterator	iter = devices.begin();
	int 							menuID = 1;

	while (iter != devices.end())
	{
		::AppendMenu (menu, MF_STRING, menuID, iter->second.c_str ());

		if (gCfg.fDeviceType == iter->first)
		{
			::CheckMenuItem (menu, menuID, MF_BYCOMMAND | MF_CHECKED);
		}

		++iter;
		++menuID;
	}

	UINT	result = ::PrvClickCommon (lpDIS, dlg, menu);

	if (result != 0)
	{
		gCfg.fDeviceType = devices [result - 1].first;
		::PrvSetDevicePopupText (dlg, gCfg.fDeviceType);

		::PrvCheckSkinPopup (dlg);
	}
}


void PrvClickSkinPopup (LPDRAWITEMSTRUCT lpDIS, HWND dlg)
{
	HMENU	menu = ::CreatePopupMenu ();

	SkinNameList	skins;
	::SkinGetSkinNames (gCfg.fDeviceType, skins);

	int 					menuID = 1;
	SkinNameList::iterator	iter = skins.begin ();
	while (iter != skins.end())
	{
		::AppendMenu (menu, MF_STRING, menuID, iter->c_str ());

		if (gSkinChoices[gCfg.fDeviceType] == *iter)
		{
			::CheckMenuItem (menu, menuID, MF_BYCOMMAND | MF_CHECKED);
		}

		++iter;
		++menuID;
	}

	UINT	result = ::PrvClickCommon (lpDIS, dlg, menu);

	if (result != 0)
	{
		gSkinChoices[gCfg.fDeviceType] = skins [result - 1];

		::PrvSetSkinPopupText (dlg, gSkinChoices[gCfg.fDeviceType]);
	}
}


void PrvClickRAMSizePopup (LPDRAWITEMSTRUCT lpDIS, HWND dlg)
{
	HMENU	menu = ::CreatePopupMenu ();

	MemoryTextList	sizes;
	::GetMemoryTextList (sizes);

	MemoryTextList::iterator	iter = sizes.begin();
	int							menuID = 1;
	while (iter != sizes.end())
	{
		::AppendMenu (menu, MF_STRING, menuID, iter->second.c_str ());

		if (gCfg.fRAMSize == iter->first)
		{
			::CheckMenuItem (menu, menuID, MF_BYCOMMAND | MF_CHECKED);
		}

		++iter;
		++menuID;
	}

	UINT	result = ::PrvClickCommon (lpDIS, dlg, menu);

	if (result != 0)
	{
		gCfg.fRAMSize = sizes [result - 1].first;
		::PrvSetRAMSizePopupText (dlg, gCfg.fRAMSize);
	}
}


UINT PrvClickCommon (LPDRAWITEMSTRUCT lpDIS, HWND dlg, HMENU menu)
{
	RECT	buttonBounds;
	::GetWindowRect (lpDIS->hwndItem, &buttonBounds);

	UINT	result = ::TrackPopupMenu (menu, TPM_NONOTIFY | TPM_RETURNCMD,
									buttonBounds.left, buttonBounds.bottom, 0, dlg, NULL);

	::DestroyMenu (menu);

	return result;
}


void PrvCheckSkinPopup (HWND hwnd)
{
	// The device menu was just changed; update the skin button text.

	::PrvSetSkinPopupText (hwnd, gSkinChoices[gCfg.fDeviceType]);
}


void PrvSetDevicePopupText (HWND hwnd, DeviceType dt)
{
	HWND dlgItem = ::GetDlgItem (hwnd, IDC_DEVICE);
	::SetWindowText (dlgItem, ::PrvGetDeviceText (dt).c_str ());
}


void PrvSetSkinPopupText (HWND hwnd, const string& skinName)
{
	HWND dlgItem = ::GetDlgItem (hwnd, IDC_SKIN);
	::SetWindowText (dlgItem, skinName.c_str ());
}


void PrvSetRAMSizePopupText (HWND hwnd, RAMSizeType ramSize)
{
	HWND dlgItem = ::GetDlgItem (hwnd, IDC_RAMSIZE);
	::SetWindowText (dlgItem, ::PrvGetRAMSizeText (ramSize).c_str ());
}


void PrvSetROMFileButtonText (HWND hwnd, const EmFileRef& romFile)
{
	HWND dlgItem = ::GetDlgItem (hwnd, IDC_ROMFILE);
	::SetWindowText (dlgItem, romFile.GetFullPath ().c_str ());
}


string PrvGetDeviceText (DeviceType dt)
{
	string result;

	if (dt == kDevicePilot5000)
		dt = kDevicePilot1000;
	else if (dt == kDevicePalmPilotProfessional)
		dt = kDevicePalmPilotPersonal;

	EmDeviceMenuItemList			devices = EmDevice::GetMenuItemList ();
	EmDeviceMenuItemList::iterator	iter = devices.begin();

	while (iter != devices.end())
	{
		if (iter->first == dt)
		{
			result = iter->second;
			break;
		}

		++iter;
	}

	return result;
}


string PrvGetRAMSizeText (RAMSizeType ramSize)
{
	string result;

	MemoryTextList	sizes;
	::GetMemoryTextList (sizes);

	MemoryTextList::iterator	iter = sizes.begin();
	while (iter != sizes.end())
	{
		if (iter->first == ramSize)
		{
			result = iter->second;
			break;
		}

		++iter;
	}

	return result;
}
