/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmJPEGWin.h"

#include "Platform.h"			// AllocateMemory


// Private EmJPEGDecompressSink that writes the output as s BMP.

class EmJPEGDecompressBMPMemSink : public EmJPEGDecompressSink
{
	public:
								EmJPEGDecompressBMPMemSink	(void);
		virtual					~EmJPEGDecompressBMPMemSink	(void);

		virtual void			post_jpeg_read_header	(j_decompress_ptr cinfo);

		virtual void			start_output	(j_decompress_ptr cinfo);
		virtual void			put_pixel_rows	(j_decompress_ptr cinfo,
												 JDIMENSION rows_supplied);
		virtual void			finish_output	(j_decompress_ptr cinfo);

	public:
		BITMAPINFO*				fBitmap;

	private:
		JDIMENSION				fDataWidth;		/* JSAMPLEs per row */
		JDIMENSION				fRowWidth;		/* physical width of one row in the BMP file */
		int						fPadBytes;		/* number of padding bytes needed per row */
		JDIMENSION				fCurOutputRow;	/* next row# to write to virtual array */
};


/***********************************************************************
 *
 * FUNCTION:	JPEGToBMP
 *
 * DESCRIPTION:	Convert a JPEG to a BMP (that is, a graphic image
 *				described by a BITMAPINFO).
 *
 * PARAMETERS:	image - pointer to the compressed JPEG data
 *				imageLen - length of the compressed JPEG data
 *
 * RETURNED:	Pointer to a BITMAPINFO expressing the BMP.  The
 *				buffer was created with Platform::AllocateMemory; the
 *				caller needs to dispose of it with DisposeMemory.
 *
 ***********************************************************************/

BITMAPINFO* JPEGToBMP (const void* image, long imageLen)
{
	EmJPEGDecompressMemSource	src_data (image, imageLen);
	EmJPEGDecompressBMPMemSink	dest_data;

	::ConvertJPEG (src_data, dest_data);

	return dest_data.fBitmap;
}


/***********************************************************************
 *
 * FUNCTION:	JPEGToBMP
 *
 * DESCRIPTION:	Convert a JPEG to a BMP (that is, a graphic image
 *				described by a BITMAPINFO).
 *
 * PARAMETERS:	stream - stream that provides the JPEG data
 *
 * RETURNED:	Pointer to a BITMAPINFO expressing the BMP.  The
 *				buffer was created with Platform::AllocateMemory; the
 *				caller needs to dispose of it with DisposeMemory.
 *
 ***********************************************************************/

BITMAPINFO* JPEGToBMP (EmStream& stream)
{
	EmJPEGDecompressStreamSource	src_data (stream);
	EmJPEGDecompressBMPMemSink		dest_data;

	::ConvertJPEG (src_data, dest_data);

	return dest_data.fBitmap;
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmJPEGDecompressBMPMemSink c'tor
 *
 * DESCRIPTION:	Create the object.  Initialize all data members.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmJPEGDecompressBMPMemSink::EmJPEGDecompressBMPMemSink (void) :
	EmJPEGDecompressSink (),
	fBitmap (NULL),
	fDataWidth (0),
	fRowWidth (0),
	fPadBytes (0),
	fCurOutputRow (0)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmJPEGDecompressBMPMemSink d'tor
 *
 * DESCRIPTION:	Destroy the object.  Delete all data members.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmJPEGDecompressBMPMemSink::~EmJPEGDecompressBMPMemSink (void)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmJPEGDecompressBMPMemSink::post_jpeg_read_header
 *
 * DESCRIPTION:	Called by the JPEG decompression engine after the JPEG
 *				header has been read.  Initializes some of our data
 *				members according to what was read.
 *
 * PARAMETERS:	cinfo - data block describing the decompression process.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void
EmJPEGDecompressBMPMemSink::post_jpeg_read_header (j_decompress_ptr cinfo)
{
	// Calculate output image dimensions so we can allocate space

	jpeg_calc_output_dimensions (cinfo);

	// Determine width of rows in the BMP file (padded to 4-byte boundary).

	fDataWidth = cinfo->output_width * cinfo->output_components;
	fRowWidth = (fDataWidth + 3) & ~0x03;
	fPadBytes = fRowWidth - fDataWidth;

	// Prepare for write pass

	fCurOutputRow = 0;

	// Create decompressor output buffer.

	fBuffer = (*cinfo->mem->alloc_sarray)
		((j_common_ptr) cinfo, JPOOL_IMAGE, fDataWidth, 1);
	fBufferHeight = 1;

	long bitmap_size = fRowWidth * cinfo->output_height + offsetof (BITMAPINFO, bmiColors);
	fBitmap = (BITMAPINFO*) Platform::AllocateMemory (bitmap_size);
}


/***********************************************************************
 *
 * FUNCTION:	EmJPEGDecompressBMPMemSink::start_output
 *
 * DESCRIPTION:	Write any initial part of the output file.
 *
 * PARAMETERS:	cinfo - data block describing the decompression process.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmJPEGDecompressBMPMemSink::start_output (j_decompress_ptr cinfo)
{
	fBitmap->bmiHeader.biSize			= sizeof (BITMAPINFOHEADER);
	fBitmap->bmiHeader.biWidth			= cinfo->output_width;
	fBitmap->bmiHeader.biHeight			= cinfo->output_height;
	fBitmap->bmiHeader.biPlanes			= 1;
	fBitmap->bmiHeader.biBitCount		= 24;
	fBitmap->bmiHeader.biCompression	= BI_RGB;
	fBitmap->bmiHeader.biSizeImage		= 0;
	fBitmap->bmiHeader.biXPelsPerMeter	= 0;
	fBitmap->bmiHeader.biYPelsPerMeter	= 0;
	fBitmap->bmiHeader.biClrUsed		= 0;
	fBitmap->bmiHeader.biClrImportant	= 0;
}


/***********************************************************************
 *
 * FUNCTION:	EmJPEGDecompressBMPMemSink::put_pixel_rows
 *
 * DESCRIPTION:	Write some pixel data.
 *
 * PARAMETERS:	cinfo - data block describing the decompression process.
 *				JDIMENSION - number of rows.  This parameter is really
 *					just a holdover from the JPEG sample code; in our
 *					implementation, it's always 1.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmJPEGDecompressBMPMemSink::put_pixel_rows (j_decompress_ptr cinfo, JDIMENSION)
{
	int		line	= cinfo->output_height - fCurOutputRow - 1;
	char*	outptr	= ((char*) fBitmap) +
						fBitmap->bmiHeader.biSize +
						line * fRowWidth;
	fCurOutputRow++;

	// Transfer data.  Note destination values must be in BGR order
	// (even though Microsoft's own documents say the opposite).

	JSAMPROW	inptr = fBuffer[0];

	for (JDIMENSION col = 0; col < cinfo->output_width; ++col)
	{
		outptr[2] = *inptr++;
		outptr[1] = *inptr++;
		outptr[0] = *inptr++;
		outptr += 3;
	}

	// Zero out the pad bytes.

	int pad = fPadBytes;
	while (--pad >= 0)
		*outptr++ = 0;
}


/***********************************************************************
 *
 * FUNCTION:	EmJPEGDecompressBMPMemSink::finish_output
 *
 * DESCRIPTION:	Write any data that needs to be written after all the
 *				pixel rows have been processed.
 *
 * PARAMETERS:	cinfo - data block describing the decompression process.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmJPEGDecompressBMPMemSink::finish_output (j_decompress_ptr /* cinfo */)
{
}
