/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Hyperlink.h"

#include "Emulator.h"			// gInstance

#include "resource.h"			// IDC_FINGER


// ======================================================================
//	Globals and constants
// ======================================================================


// ===========================================================================
//		 SubclassedWindow
// ===========================================================================

/***********************************************************************
 *
 * FUNCTION:	SubclassedWindow constructor
 *
 * DESCRIPTION:	Remember the original window procedure and install
 *				ours in its place.
 *
 * PARAMETERS:	hwnd - window to subclass
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

SubclassedWindow::SubclassedWindow(HWND hwnd) :
	Hwnd(hwnd)
{
	fOldWndProc = (WNDPROC) ::GetWindowLong (Hwnd, GWL_WNDPROC);
	::SetWindowLong (Hwnd, GWL_WNDPROC, (long) &StaticWndProc);
	::SetWindowLong (Hwnd, GWL_USERDATA, (long) this);
}


/***********************************************************************
 *
 * FUNCTION:	SubclassedWindow destructor
 *
 * DESCRIPTION:	Restore the original window procedure.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

SubclassedWindow::~SubclassedWindow()
{
	::SetWindowLong (Hwnd, GWL_WNDPROC, (long) fOldWndProc);
}


/***********************************************************************
 *
 * FUNCTION:	SubclassedWindow::StaticWndProc
 *
 * DESCRIPTION:	Installed window procedure.  Retrieves the pointer
 *				to the "SubclassedWindow" object associated with this
 *				window and calls its (virtual) WndProc method.
 *
 * PARAMETERS:	Standard window procedure parameters
 *
 * RETURNED:	Standard window procedure result
 *
 ***********************************************************************/

LRESULT CALLBACK SubclassedWindow::StaticWndProc (HWND hwnd, UINT msg,
												  WPARAM wparam, LPARAM lparam)
{
	SubclassedWindow*	This = reinterpret_cast<SubclassedWindow*> (
										::GetWindowLong (hwnd, GWL_USERDATA));
	LRESULT 			r = This->WndProc (hwnd, msg, wparam, lparam);

	if (msg == WM_NCDESTROY)		// last message this window will get
	{
		::SetWindowLong (This->Hwnd, GWL_WNDPROC, (long) This->fOldWndProc);
		delete This;
	}

	return r;
}


// ===========================================================================
//		 StaticHyperlink
// ===========================================================================

/***********************************************************************
 *
 * FUNCTION:	StaticHyperlink constructor
 *
 * DESCRIPTION:	Initialize our member data.
 *
 * PARAMETERS:	hwnd - window to subclass.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

StaticHyperlink::StaticHyperlink (HWND hwnd) :
	SubclassedWindow(hwnd)
{
	font = NULL;
	finger = ::LoadCursor (gInstance, MAKEINTRESOURCE (IDC_FINGER));
	color = RGB (0, 0, 255);
}


/***********************************************************************
 *
 * FUNCTION:	StaticHyperlink destructor
 *
 * DESCRIPTION:	Release our resources.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

StaticHyperlink::~StaticHyperlink()
{
	if (font)
	{
		::DeleteObject (font);
	}
}


/***********************************************************************
 *
 * FUNCTION:	StaticHyperlink::OnCtlColorStatic
 *
 * DESCRIPTION:	Handle WM_CTLCOLORSTATIC messages sent to our window.
 *				For Hyperlinks, set the color appropriate for selected
 *				and unselected links.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

HBRUSH StaticHyperlink::OnCtlColorStatic (HWND /*parent*/, HDC dc, HWND /*control*/, int /*type*/)
{
	HBRUSH	hbr = NULL;

	if ((GetWindowLong (Hwnd, GWL_STYLE) & 0xFF) <= SS_RIGHT)
	{
		if (font == NULL)
		{
			LOGFONT lf;

			GetObject ((HFONT) SendMessage (Hwnd, WM_GETFONT, 0, 0), sizeof (lf), &lf);

			lf.lfUnderline = TRUE;
			font = CreateFontIndirect (&lf);
		}

		SelectObject (dc, font);
		SetTextColor (dc, color);
		SetBkMode (dc, TRANSPARENT);
		hbr = (HBRUSH) GetStockObject (HOLLOW_BRUSH);
	}

	return hbr;
}


/***********************************************************************
 *
 * FUNCTION:	StaticHyperlink::OnLButtonDown
 *
 * DESCRIPTION:	Handle WM_LBUTTONDOWN messages sent to our window.
 *				For Hyperlinks, launch the associated URL by calling
 *				ShellExecute.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void StaticHyperlink::OnLButtonDown (HWND hwnd, BOOL doubleclick, int x, int y, UINT flags)
{
	UNUSED_PARAM (hwnd);
	UNUSED_PARAM (doubleclick);
	UNUSED_PARAM (x);
	UNUSED_PARAM (y);
	UNUSED_PARAM (flags);

	// Get the URL.

	char	buf[256];
	::GetWindowText (Hwnd, buf, sizeof (buf));

	// Strip off the leading and trailing elbow-brackets.

	size_t	len = strlen (buf);
	memcpy (buf, &buf[1], len - 2);
	buf[len - 2] = 0;

	HCURSOR cur = ::GetCursor ();
	::SetCursor (::LoadCursor (NULL, IDC_WAIT));

	HINSTANCE	r = ::ShellExecute (NULL, "open", buf, NULL, NULL, SW_SHOWNORMAL);

	::SetCursor (cur);

	if ((UINT) r > 32)
	{
		color = RGB (128, 0, 128);
		::InvalidateRect (Hwnd, NULL, FALSE);
	}
}


/***********************************************************************
 *
 * FUNCTION:	StaticHyperlink::OnNcHitTest
 *
 * DESCRIPTION:	Handle WM_NCHITTEST messages sent to our window.
 *				For some reason, always returning HTCLIENT is good.
 *				I guess static text items always return some sort of
 *				"nothing hit" code otherwise.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

UINT StaticHyperlink::OnNcHitTest (HWND hwnd, int x, int y)
{
	UNUSED_PARAM (hwnd);
	UNUSED_PARAM (x);
	UNUSED_PARAM (y);

	return HTCLIENT;
}


/***********************************************************************
 *
 * FUNCTION:	StaticHyperlink::OnSetCursor
 *
 * DESCRIPTION:	Handle WM_SETCURSOR messages sent to our window.
 *				For Hyperlinks, set the cursor to the de facto
 *				standard that browsers use when the cursor is over
 *				an active hyperlink.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

BOOL StaticHyperlink::OnSetCursor (HWND hwnd, HWND hwndCursor, UINT codeHitTest, UINT msg)
{
	UNUSED_PARAM (hwnd);
	UNUSED_PARAM (hwndCursor);
	UNUSED_PARAM (codeHitTest);
	UNUSED_PARAM (msg);

	::SetCursor (finger);

	return TRUE;
}


/***********************************************************************
 *
 * FUNCTION:	StaticHyperlink::WndProc
 *
 * DESCRIPTION:	Respond to the Windows messages appropriate for turning
 *				a static text dialog item into an active URL Hyperlink.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

LRESULT StaticHyperlink::WndProc (HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch (msg)
	{
		case WM_CTLCOLORSTATIC:
			return HANDLE_WM_CTLCOLORSTATIC (hwnd, wparam, lparam, OnCtlColorStatic);

		case WM_LBUTTONDOWN:
			return HANDLE_WM_LBUTTONDOWN (hwnd, wparam, lparam, OnLButtonDown);

		case WM_NCHITTEST:
			return HANDLE_WM_NCHITTEST (hwnd, wparam, lparam, OnNcHitTest);

		case WM_SETCURSOR:
			return HANDLE_WM_SETCURSOR (hwnd, wparam, lparam, OnSetCursor);

		default:
			return ::CallWindowProc (fOldWndProc, hwnd, msg, wparam, lparam);
	}

	return 0;
}

