/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "LCD.h"

#include "Application.h"		// DisplayWndProc
#include "CPU_REG.h"			// Screen::LCDIsOn, Screen::BacklightIsOn, Screen::GetBits, etc.
#include "CPU_MT.h" 			// CPU_ functions
#include "EmDlgWin.h"			// CenterWindow, CenterDlgProc
#include "EmJPEGWin.h"			// JPEGToBMP
#include "Emulator.h"			// gInstance
#include "Logging.h"			// LogAppendMsg
#include "Quantizer.h"			// CQuantizer
#include "Skins.h"				// SkinElementType, SkinGetElementRect, etc.
#include "Strings.r.h"			// kStr_AppName

#include "resource.h"			// IDI_EMULATOR, etc.

// ======================================================================
//	Globals and constants
// ======================================================================

const char* 	kClassName	= "PalmOSEmulatorDisplay";
const char* 	RT_IMAGE	= "IMAGE";


HWND			gEmulatorWindow;

BITMAPINFO* 	gCaseBitmap;
RGBQUAD*		gCaseColors;
int 			gNumCaseColors;

HPALETTE		gCachedPalette;
HBITMAP 		gCachedCase;

BOOL			gPenTracking;
BOOL			gButtonTracking;
SkinElementType gCurrentButton;

BOOL			gNeedWindowReset;
BOOL			gNeedWindowInvalidate;


// ======================================================================
//	Private functions
// ======================================================================

static void 	PrvSetColorRamp (RGBQUAD colors[], int numColors, RGBType* palette);
static void 	PrvPaintCase (HDC destDC, const Screen::BufferInfo& info);
static void 	PrvPaintLCD (HDC destDC, const Screen::BufferInfo& info);
static void 	PrvPaintButtonFrames (HDC destDC);
static void 	PrvPaintScreen (HDC destDC, BOOL drawCase, BOOL always);
static HPALETTE PrvCreateLCDPalette (HDC destDC, const Screen::BufferInfo& info);
static HBITMAP	PrvCreateLCDBitmap (HDC destDC, const Screen::BufferInfo& info);
static void 	PrvLoadResourceBitmap (HINSTANCE hInstance,
										BITMAPINFO** bitmap, RGBQUAD** colors, int* numColors,
										HRGN* outline, RECT* bounds);
static BITMAPINFO*	PrvGetDefaultBitmap (void);

static HPALETTE PrvCreateDIBPalette (LPBITMAPINFO dib, void* bits);
static void 	PrvGetColors (LPBITMAPINFO dib, void* bits, RGBQUAD** colors, int* numColors);
static HPALETTE PrvGetPalette (const Screen::BufferInfo& info);
static BOOL 	PrvIsPaletteDevice (HDC hDC);
static long 	PrvGetRowBytes (LPBITMAPINFO dib);
static long 	PrvGetRowBytes (long width, long bpp);
static int		PrvGetNumColors (LPBITMAPINFO dib);
static BYTE*	PrvGetBits (LPBITMAPINFO dib);
static RECT 	PrvGetRECT (const RectangleType& r);

static HRGN 	PrvCreateOutlineRegion (LPBITMAPINFO dib);
static void 	PrvConvert24To1BPP (void* dstPtr, void* srcPtr, int width, int height);
static void		PrvConvert8To1BPP (void* dstPtr, void* srcPtr, int width, int height, RGBQUAD* colors);
static void 	PrvMakeMask (void* dstPtr, void* srcPtr, int width, int height);
static HRGN 	PrvBitmapToRegion (void* srcPtr, int width, int height);
static void 	PrvAddToRegion (HRGN& region, int yy, int x1, int x2);

static void		PrvGetLCDBounds (RectangleType&);
static void		PrvGetLCDBounds (AbsRectType&);


#define WIDTH(r)	((r).right - (r).left)
#define HEIGHT(r)	((r).bottom - (r).top)


/***********************************************************************
 *
 * FUNCTION:	PrvSetColorRamp
 *
 * DESCRIPTION:	Helper routine for LCD_ScreenShot.	Fills in a color
 *				table with appropriate gray values.
 *
 * PARAMETERS:	colors - RGBQUAD array to receive the colors.
 *
 *				numColors - number of colors to write into "colors".
 *
 *				foreground, background - starting and ending colors of
 *				the ramp to create.
 *
 * RETURNED:	New colors are written into "colors".
 *
 ***********************************************************************/

static void PrvSetColorRamp (RGBQUAD colors[], int numColors, 
				   RGBType* palette)
{
	for (int color = 0; color < numColors; color++)
	{
		colors[color].rgbRed		= palette[color].fRed;
		colors[color].rgbGreen		= palette[color].fGreen;
		colors[color].rgbBlue		= palette[color].fBlue; 
		colors[color].rgbReserved	= 0;
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvPaintScreen
 *
 * DESCRIPTION:	Called by LCD_Refresh and LCD_Paint to do the real
 *				dirty work.
 *
 * PARAMETERS:	dc - HDC to draw to.
 *
 *				always - true to redraw the whole screen, false to draw
 *				just the parts that were changed by the emulated process
 *				since the last time we updated.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void PrvPaintCase (HDC destDC, const Screen::BufferInfo& info)
{
	int 	width	= gCaseBitmap->bmiHeader.biWidth;
	int 	height	= gCaseBitmap->bmiHeader.biHeight;
	void*	bits	= ::PrvGetBits (gCaseBitmap);

#if 0	// Not very fast in 8-bit mode
	int result = ::StretchDIBits (
		destDC, 		// handle to device context
		0, 0,			// x/y-coordinate of upper-left corner of dest. rectangle
		width, height,	// width/height of destination rectangle
		0, 0,			// x/y-coordinate of upper-left corner of source rectangle
		width, height,	// width/height of source rectangle
		bits,			// address of bitmap bits
		gCaseBitmap,	// address of bitmap data
		DIB_RGB_COLORS, // usage flags
		SRCCOPY);		// raster operation code
#endif

#if 0	// This isn't any better
	int result = ::SetDIBitsToDevice (
		destDC, 		// handle to device context
		0, 0,			// x/y-coordinate of upper-left corner of dest. rect.
		width, height,	// source rectangle width/height
		0, 0,			// x/y-coordinate of lower-left corner of source rect.
		0,				// first scan line in array
		height, 		// number of scan lines
		bits,			// address of array with DIB bits
		gCaseBitmap,	// address of structure with bitmap info.
		DIB_RGB_COLORS);// RGB or palette indexes
#endif

	BOOL	cacheImage = !gCachedCase;

	if (cacheImage)
	{
		gCachedCase = ::CreateCompatibleBitmap (destDC, width, height);
	}

	HDC 	srcDC = ::CreateCompatibleDC (destDC);
	HBITMAP oldBitmap = SelectBitmap (srcDC, gCachedCase);

	if (cacheImage)
	{
		HPALETTE	pal = ::PrvGetPalette (info);
		HPALETTE	oldPal = NULL;

		if (pal)
		{
			oldPal = ::SelectPalette (srcDC, pal, FALSE);
		}

		int result = ::SetDIBitsToDevice (
			srcDC,			// handle to device context
			0, 0,			// x/y-coordinate of upper-left corner of dest. rect.
			width, height,	// source rectangle width/height
			0, 0,			// x/y-coordinate of lower-left corner of source rect.
			0,				// first scan line in array
			height, 		// number of scan lines
			bits,			// address of array with DIB bits
			gCaseBitmap,	// address of structure with bitmap info.
			DIB_RGB_COLORS);// RGB or palette indexes

		if (oldPal)
		{
			::SelectPalette (srcDC, oldPal, FALSE);
		}
	}

	BOOL	result = ::BitBlt (
		destDC, 		// handle to destination device context
		0, 0,			// y-coordinate of destination rectangle's upper-left corner
		width, height,	// width/height of destination rectangle
		srcDC,			// handle to source device context
		0, 0,			// x/y-coordinate of source rectangle's upper-left corner
		SRCCOPY);		// raster operation code

	SelectBitmap (srcDC, oldBitmap);
	::DeleteDC (srcDC);
}


void PrvPaintLCD (HDC destDC, const Screen::BufferInfo& info)
{
	// ======================================================================
	// Get the bounds of the area we'll be blitting from.
	// ======================================================================

	RECT	srcRect;
	srcRect.top 	= info.firstLine;
	srcRect.bottom	= info.lastLine;
	srcRect.left	= info.visibleBounds.left;
	srcRect.right	= info.visibleBounds.right;

	// ======================================================================
	// Get the bounds of the LCD area.
	// ======================================================================

	RectangleType	lcdRect;
	::PrvGetLCDBounds (lcdRect);
	lcdRect = ::SkinScaleDown (lcdRect);

	// ======================================================================
	// There is a moment during boot-up when the LCD width is 1024.
	// We don't really want that as the width (we only allow for 160
	// pixels in our window), so clip to 160.
	// ======================================================================

	if (WIDTH (srcRect) > lcdRect.extent.x)
	{
		srcRect.right = srcRect.left + lcdRect.extent.x;
	}

	// ======================================================================
	// Get the bounds of the area we'll be blitting to.  This is based on
	// the LCD bounds and the bounds of the LCD area that has changed.
	// ======================================================================

	// Set the top and bottom of the dest bounds.

	lcdRect.topLeft.y += info.firstLine;
	lcdRect.extent.y = info.lastLine - info.firstLine;

	// If we're in 2x mode, scale the dest bounds up.

	lcdRect = ::SkinScaleUp (lcdRect);

	// Convert the Palm OS RectangleType to a Windows RECT.

	RECT	destRect;
	destRect.left	= lcdRect.topLeft.x;
	destRect.top	= lcdRect.topLeft.y;
	destRect.right	= lcdRect.topLeft.x + lcdRect.extent.x;
	destRect.bottom = lcdRect.topLeft.y + lcdRect.extent.y;


	// ======================================================================
	// Paint the frame for the backlight.
	// ======================================================================

	if (Emulator::GetDevice().HasSED1375())
	{
		COLORREF	backlightColor = RGB(0xff, 0xff, 0xff);

		AbsRectType lcdRect;
		::PrvGetLCDBounds (lcdRect);
		lcdRect = ::SkinScaleDown (lcdRect);
		lcdRect.left	-= 2;
		lcdRect.top 	-= 2;
		lcdRect.right	+= 2;
		lcdRect.bottom	+= 2;
		lcdRect = ::SkinScaleUp (lcdRect);

		PointType	penSize = { 2, 2 };
		penSize = ::SkinScaleUp (penSize);

		HBRUSH	theBrush = (HBRUSH) ::GetStockObject (HOLLOW_BRUSH);
		HBRUSH	oldBrush = SelectBrush (destDC, theBrush);

		HPEN	thePen = ::CreatePen (PS_INSIDEFRAME, penSize.x, backlightColor);
		HPEN	oldPen = SelectPen (destDC, thePen);

		::Rectangle(destDC, lcdRect.left, 
							lcdRect.top, 
							lcdRect.right, 
							lcdRect.bottom);

		SelectPen (destDC, oldPen);
		SelectBrush (destDC, oldBrush);

		::DeleteObject (thePen);
		::DeleteObject (theBrush);
	}

	// ======================================================================
	// Create the GDI objects we'll need to copy the bitmap to the screen.
	//
	// Interestingly, caching the objects we create here and using them in
	// subsequent calls gets us nothing.  So, for simplicity's sake, we'll
	// keep the straightforward implementation of creating the objects
	// every time.
	// ======================================================================

	HDC 		lcdDC		= ::CreateCompatibleDC (destDC);
	HBITMAP 	lcdBM		= ::PrvCreateLCDBitmap (destDC, info);
	HPALETTE	lcdPalette	= ::PrvCreateLCDPalette (destDC, info);
	HPALETTE	oldPalette	= ::SelectPalette (lcdDC, lcdPalette, FALSE);
	HBITMAP 	oldBitmap	= SelectBitmap (lcdDC, lcdBM);

	// ======================================================================
	// Blit the image to the screen.
	// ======================================================================

	::StretchBlt (	destDC, 			// handle to destination device context
					destRect.left,		// x-coordinate of upper-left corner of dest. rect.
					destRect.top,		// y-coordinate of upper-left corner of dest. rect.
					WIDTH (destRect),	// width of destination rectangle
					HEIGHT (destRect),	// height of destination rectangle
					lcdDC,				// handle to source device context
					srcRect.left,		// x-coordinate of upper-left corner of source rectangle
					srcRect.top,		// y-coordinate of upper-left corner of source rectangle
					WIDTH (srcRect),	// width of source rectangle
					HEIGHT (srcRect),	// height of source rectangle
					SRCCOPY);

	// ======================================================================
	// Clean up.
	// ======================================================================

	::SelectPalette (lcdDC, oldPalette, FALSE);
	SelectBitmap (lcdDC, oldBitmap);

	::DeleteObject (lcdPalette);
	::DeleteObject (lcdBM);
	::DeleteDC (lcdDC);
}


void PrvPaintButtonFrames (HDC destDC)
{
	// Draw frames around the buttons.	This is handy when determining their
	// locations initially.

	HPEN	oldPen = SelectPen (destDC, ::GetStockObject (BLACK_PEN));
	HBRUSH	oldBrush = SelectBrush (destDC, ::GetStockObject (NULL_BRUSH));

	int	index = 0;
	SkinElementType	type;
	AbsRectType		bounds;

	while (::SkinGetElementInfo (index, type, bounds))
	{
		::Rectangle (destDC, bounds.left, bounds.top, bounds.right, bounds.bottom);

		++index;
	}

	SelectBrush (destDC, oldBrush);
	SelectPen (destDC, oldPen);
}


void PrvPaintScreen (HDC destDC, BOOL drawCase, BOOL always)
{
	// Get information about the screen.

	Screen::BufferInfo	info;
	Bool	drawLCD = Screen::GetBits (info, !always);

	// !!! At this point, we *could* restart the CPU thread. TBD.


	// If the on/off state of the LCD changed from on to off, make sure we
	// draw the case.  That way, we replace the LCD bits with the background.

	static	oldLCDOn;
	if (oldLCDOn && !info.lcdOn)
	{
		drawCase = true;
	}
	oldLCDOn = info.lcdOn;


	// If we need to draw anything, get the palette we'll be using.

	HPALETTE	destPal = NULL;
	HPALETTE	oldDestPal = NULL;

	if (drawLCD || drawCase)
	{
		destPal = ::PrvGetPalette (info);

		if (destPal)
		{
			oldDestPal = ::SelectPalette (destDC, destPal, FALSE);
			::RealizePalette (destDC);
		}
	}


	// Draw the case.

	if (drawCase)
		::PrvPaintCase (destDC, info);


	// Draw the LCD area.

	if (drawLCD && info.lcdOn)
		::PrvPaintLCD (destDC, info);


#if 0
	// Draw frames around the buttons.	This is handy when determining their
	// locations initially.

	if (drawCase)
		::PrvPaintButtonFrames (destDC)
#endif


	// Clean up and exit.

	if (oldDestPal)
	{
		::SelectPalette (destDC, oldDestPal, FALSE);
		::RealizePalette (destDC);
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvCreateLCDPalette
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

HPALETTE PrvCreateLCDPalette (HDC destDC, const Screen::BufferInfo& info)
{
	// Convert the color information in "info" into an
	// array of RGBQUADs.

	int 		numColors = 1 << info.lcdDepth;
	RGBQUAD*	colors = (RGBQUAD*) _alloca (numColors * sizeof (RGBQUAD));

	::PrvSetColorRamp (colors, numColors, info.myCLUT);

	// Create a LOGPALETTE of the right size and fill in the fields.

	LOGPALETTE* logPal = (LOGPALETTE*) _alloca (sizeof (LOGPALETTE) + sizeof (PALETTEENTRY) * numColors);
	logPal->palVersion		= 0x300;
	logPal->palNumEntries	= numColors;

	for (int ii = 0; ii < numColors; ++ii)
	{
		logPal->palPalEntry[ii].peRed	= colors[ii].rgbRed;
		logPal->palPalEntry[ii].peGreen = colors[ii].rgbGreen;
		logPal->palPalEntry[ii].peBlue	= colors[ii].rgbBlue;
		logPal->palPalEntry[ii].peFlags = 0;
	}

	// Create the palette object from the LOGPALETTE.

	HPALETTE	lcdPalette = ::CreatePalette (logPal);

	return lcdPalette;
}


/***********************************************************************
 *
 * FUNCTION:	PrvCreateLCDBitmap
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

HBITMAP PrvCreateLCDBitmap (HDC destDC, const Screen::BufferInfo& info)
{
	HBITMAP lcdBM = NULL;

	// Fill in a BITMAPINFO structure accordingly so that we can turn
	// the LCD bits into an HBITMAP.

	int numColors = 1 << info.lcdDepth;

	BITMAPINFO* bmi = (BITMAPINFO*) _alloca (sizeof (BITMAPINFOHEADER) +
											numColors * sizeof (RGBQUAD));
	::ZeroMemory (bmi, sizeof (BITMAPINFOHEADER));
	bmi->bmiHeader.biSize		= sizeof (BITMAPINFOHEADER);
	bmi->bmiHeader.biWidth		= info.myRowBytes * 8 / info.myDepth;
	bmi->bmiHeader.biHeight 	= -info.height;
	bmi->bmiHeader.biPlanes 	= 1;
	bmi->bmiHeader.biBitCount	= info.myDepth;
	bmi->bmiHeader.biClrUsed	= numColors;

	::PrvSetColorRamp (bmi->bmiColors, numColors, info.myCLUT);

	// Create the DIB section.	The last two parameters to CreateDIBSection
	// are hSection and dwOffset; by setting them to NULL/0, we tell Windows
	// to allocate the memory for us.

	VOID*	bits;
	lcdBM = ::CreateDIBSection (destDC, bmi, DIB_RGB_COLORS, &bits, NULL, 0);

	if (!lcdBM || !bits)
	{
		// If we couldn't get that, just return.  I'm not sure why CreateDIBSection
		// would fail, but I've seen it happen.  It usually happens when another
		// copy of the emulator contacts this one in order to download the ROM.
		// Perhaps the intense serial port activity is getting in the way.
		// Whatever.  But subsequent attempts to call CreateDIBSection appear to
		// work.  GetLastError is returning zero.

		DWORD	error = ::GetLastError ();

		return lcdBM;
	}

	// Copy our bits into it. They should already be in the right format
	// from the Screen::GetBits call.
	//
	// Note: modifying GetBits to copy directly into the "bits" buffer (instead
	// of using an intermediate buffer) doesn't get us anything (according
	// to timing tests).

	memcpy (((char*) bits) + info.firstLine * info.myRowBytes,
			((char*) info.myBuffer) + info.firstLine * info.myRowBytes,
			(info.lastLine - info.firstLine) * info.myRowBytes);

	return lcdBM;
}


/***********************************************************************
 *
 * FUNCTION:    PrvPrefsChanged
 *
 * DESCRIPTION: If any preference relevent to our display has changed,
 *				set a flag remembering it.  We'll respond to that flag
 *				later in some idle-time function.  We don't respond
 *				to it now in case several preferences change all at
 *				once; we don't want to respond to each one individually.
 *
 * PARAMETERS:  key - the key of the preference that changed.
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

static void PrvPrefsChanged (PrefKeyType key)
{
	if ((strcmp (key, kPrefKeyScale) == 0) || (strcmp (key, kPrefKeySkins) == 0))
	{
		gNeedWindowReset = true;
	}
	else if (strcmp (key, kPrefKeyBackgroundColor) == 0)
	{
		gNeedWindowInvalidate = true;
	}
}


/***********************************************************************
 *
 * FUNCTION:	LCD_Init
 *
 * DESCRIPTION:	Called when the application starts up.	Loads in any
 *				resources, sets up color palettes, and creates the
 *				emulator LCD window.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_Init (void)
{
	// Initialize the flag used for emulating sounds based on our preference settings.
	// !!! TBD

//	::CPU_setsounds (gSettings.Sounds != 0);

	// Create the window class for the emulator window.

	WNDCLASS	wc;
	::ZeroMemory (&wc, sizeof (wc));

	wc.style			= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc		= &::DisplayWndProc;
	wc.hInstance		= gInstance;
	wc.hIcon			= ::LoadIcon (gInstance, MAKEINTRESOURCE (IDI_EMULATOR));
	wc.hCursor			= ::LoadCursor (NULL, IDC_ARROW);
	wc.lpszClassName	= kClassName;

	::RegisterClass (&wc);

	// Create the emulator window and show it.

	string	app (Platform::GetString (kStr_AppName));

	Preference<PointType>	pref (kPrefKeyWindowLocation);
	PointType				loc = {0, 0};

	if (pref.Loaded ())
	{
		loc = *pref;
	}

	// Create the window.  Use WS_POPUP to create a captionless window.
	// Otherwise, we'd have to remove the caption style bit in LCD_ResetWindow.
	// See MSDN article "Win32 Window Hierarchy and Styles".

	gEmulatorWindow = ::CreateWindow (	kClassName, "Palm Emulator",
										WS_SYSMENU | WS_MINIMIZEBOX | WS_POPUP,
										loc.x, loc.y,
										CW_USEDEFAULT, CW_USEDEFAULT,
										NULL, NULL,			// parent, menu
										gInstance, NULL);	// lpParam

	if (!pref.Loaded ())
	{
		::CenterWindow (gEmulatorWindow);
	}

	// Ensure the window is on the screen.

	RECT	wr;
	::GetWindowRect (gEmulatorWindow, &wr);

	AbsRectType	r = { wr.left, wr.top, wr.right, wr.bottom };
	if (Platform::PinToScreen (r))
	{
		::SetWindowPos (gEmulatorWindow, NULL, r.left, r.top,
					-1, -1, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
	}

	::ShowWindow (gEmulatorWindow, SW_SHOWNORMAL);
	::SetForegroundWindow (gEmulatorWindow);

	gPrefs->AddNotification (PrvPrefsChanged, kPrefKeySkins);
	gPrefs->AddNotification (PrvPrefsChanged, kPrefKeyScale);
	gPrefs->AddNotification (PrvPrefsChanged, kPrefKeyBackgroundColor);
}


/***********************************************************************
 *
 * FUNCTION:	LCD_Quit
 *
 * DESCRIPTION:	Called when the application is about to end.  Releases
 *				our resources.	Doesn't need to delete the window, as
 *				its deletion was what got us here in the first place.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_Quit (void)
{
	// No longer need notification.

	gPrefs->RemoveNotification (PrvPrefsChanged);

	// Release our resources before quitting.

	Platform::DisposeMemory (gCaseBitmap);
	Platform::DisposeMemory (gCaseColors);
	::DeleteObject (gCachedCase);
	::DeleteObject (gCachedPalette);

	gCaseBitmap = NULL;
	gCaseColors = NULL;
	gCachedCase = NULL;
	gCachedPalette = NULL;
}


/***********************************************************************
 *
 * FUNCTION:	LCD_GetWindow
 *
 * DESCRIPTION:	Returns the HWND of the emulator window for any
 *				external routines that need it (such as those that need
 *				to create a new window with the LCD window as the
 *				parent).
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	HWND of the emulator LCD window.
 *
 ***********************************************************************/

HWND LCD_GetWindow (void)
{
	return gEmulatorWindow;
}


/***********************************************************************
 *
 * FUNCTION:	LCD_SetWindow
 *
 * DESCRIPTION:	Sets the HWND returned by LCD_GetWindow.  Called by the
 *				LCD's WndProc when the window is destroyed.
 *
 * PARAMETERS:	HWND of the emulator LCD window.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_SetWindow (HWND hwnd)
{
	gEmulatorWindow = hwnd;
}


/***********************************************************************
 *
 * FUNCTION:	LCD_ResetWindow
 *
 * DESCRIPTION:	Called in response to WM_CREATE messages, and when the
 *				user changes the appropriate settings in the Properties
 *				dialog box.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_ResetWindow (HWND hWnd)
{
	// Delete the old bitmap and palette.

	Platform::DisposeMemory (gCaseBitmap);
	Platform::DisposeMemory (gCaseColors);
	::DeleteObject (gCachedCase);
	::DeleteObject (gCachedPalette);

	gCaseBitmap = NULL;
	gCaseColors = NULL;
	gCachedCase = NULL;
	gCachedPalette = NULL;

	// Configure our Skin engine.

	::SkinSetSkin ();

	// Get the determined/desired bitmap.  PrvLoadResourceBitmap is a private
	// function that also extracts and returns palette information.

	HRGN	rgn;
	RECT	bounds;
	::PrvLoadResourceBitmap (gInstance, &gCaseBitmap, &gCaseColors, &gNumCaseColors, &rgn, &bounds);

	// Change the window to accomodate the settings and bitmap.

	// Get the new window size/location.
	RECT wr, cr;
	::GetWindowRect (hWnd, &wr);
	::GetClientRect (hWnd, &cr);

	// Resize the window.
	::MoveWindow (hWnd, wr.left, wr.top,
			WIDTH(bounds) + WIDTH(wr) - WIDTH(cr),
			HEIGHT(bounds) + HEIGHT(wr) - HEIGHT(cr), TRUE);

	// Set the framing region for the window.
	::SetWindowRgn (hWnd, rgn, TRUE);

	// Set a timer so that we periodically refresh the window contents.
	// Also, invalidate the window contents now (necessary?).

	::SetTimer (hWnd, 1, 1000 / 10 /*gSettings.FPS*/, NULL);
	::InvalidateRect (hWnd, NULL, FALSE);
}


/***********************************************************************
 *
 * FUNCTION:	LCD_PenEvent
 *
 * DESCRIPTION:	Called in response to WM_LBUTTONDOWN, WM_LBUTTONUP, and
 *				WM_MOUSEMOVE messages.
 *
 * PARAMETERS:	hwnd - window received in the GetMessage MSG.
 *
 *				lparam - LPARAM received in the GetMessage MSG.
 *
 *				down - true if the mouse button is down.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_PenEvent (HWND hwnd, int x, int y, Bool down)
{
	// Determine the mouse's location in 1x scale and where 0,0
	// is 0,0 on the LCD.

	PointType		where = { x, y };
	SkinElementType what = ::SkinTestPoint (where);

	// If tracking pen movement across the screen handle that.
	if (gPenTracking)
	{
		PointType	whereLCD = ::SkinWindowToTouchscreen (where);
		Hardware::PenEvent (down, whereLCD.x, whereLCD.y);

		gPenTracking = down;

		if (!gPenTracking)
		{
			::ReleaseCapture ();
		}
	}

	// If clicking on a hard button, handle that.
	else if (gButtonTracking)
	{
		gButtonTracking = down;

		if (!gButtonTracking)
		{
			::ReleaseCapture ();
			Hardware::ButtonEvent (false, gCurrentButton);
		}
	}

	// No previous mouse tracking operations are in progress.
	// Look to start something new.
	else if (down)
	{
		// If in the LCD area (including the silkscreen area)
		// start "pen tracking".

		if (what == kElement_Touchscreen)
		{
			::SetCapture (hwnd);

			PointType	whereLCD = ::SkinWindowToTouchscreen (where);
			Hardware::PenEvent (true, whereLCD.x, whereLCD.y);

			gPenTracking = TRUE;
		}

		// In a button; start button tracking.

		else
		{
			::SetCapture (hwnd);

			Hardware::ButtonEvent (true, what);

			gCurrentButton = what;
			gButtonTracking = TRUE;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	LCD_Refresh
 *
 * DESCRIPTION:	Called in response to WM_TIMER messages (to periodically
 *				update our screen).
 *
 * PARAMETERS:	hwnd - window received in the GetMessage MSG.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_Refresh (HWND hwnd)
{
	HDC dc = ::GetDC (hwnd);

	if (gNeedWindowReset)
		::LCD_ResetWindow (hwnd);
	else if (gNeedWindowInvalidate)
		::PrvPaintScreen (dc, FALSE, TRUE);
	else
		::PrvPaintScreen (dc, FALSE, FALSE);

	gNeedWindowReset = FALSE;
	gNeedWindowInvalidate = FALSE;

	::ReleaseDC (hwnd, dc);
}


/***********************************************************************
 *
 * FUNCTION:	LCD_Paint
 *
 * DESCRIPTION:	Redraws the window in response to WM_PAINT messages.
 *
 * PARAMETERS:	hwnd - window received in the GetMessage MSG.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_Paint (HWND hwnd)
{
	PAINTSTRUCT ps;
	HDC 		destDC = ::BeginPaint (hwnd, &ps);

	::PrvPaintScreen (destDC, TRUE, TRUE);

	::EndPaint (hwnd, &ps);
}


/***********************************************************************
 *
 * FUNCTION:	LCD_SelectPalette
 *
 * DESCRIPTION:	Called in response to WM_PALETTECHANGED and
 *				WM_QUERYNEWPALETTE messages.  Installs the custom
 *				palette for our window (if we're using the "realistic"
 *				display, and not the white one).
 *
 * PARAMETERS:	hwnd - window received in the GetMessage MSG.
 *
 * RETURNED:	The result of RealizePalette (the number of entries in
 *				the logical palette mapped to the system palette if the
 *				function succeeds, GDI_ERROR if it fails) or -1 if we're
 *				not in "realistic" mode.
 *
 ***********************************************************************/

int LCD_SelectPalette (HWND hwnd)
{
	// Get information about the screen.

	// !!! Problem here: getting the bits at this point could mess up
	// PrvPaintScreen when *it* gets the bits.	It could be that our
	// call here picks up on a change in the LCD bits, which could prevent
	// PrvPaintScreen from seeing the change.

	Screen::BufferInfo	info;
	Bool	drawLCD = Screen::GetBits (info, true /* incremental */);


	// Get the palette we prefer.

	HPALETTE	pal = ::PrvGetPalette (info);

	if (pal)
	{
		HDC 		dc = ::GetDC (hwnd);
		HPALETTE	op = ::SelectPalette (dc, pal, FALSE);
		UINT		n = ::RealizePalette (dc);

		if (n != GDI_ERROR)
		{
			if (n > 0)
			{
				::InvalidateRect (hwnd, NULL, FALSE);
			}
		}

		::SelectPalette (dc, op, TRUE);
		::RealizePalette (dc);

		::ReleaseDC (hwnd, dc);

		return (int) n;
	}

	return -1;
}


/***********************************************************************
 *
 * FUNCTION:	LCD_DisplayChanged
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void LCD_DisplayChanged (HWND hwnd)
{
	::DeleteObject (gCachedCase);
	gCachedCase = NULL;

	::DeleteObject (gCachedPalette);
	gCachedPalette = NULL;
}


/***********************************************************************
 *
 * FUNCTION:	LCD_ScreenShot
 *
 * DESCRIPTION:	Called to save the screen contents to a file when the
 *				user selects the "Save Screen..." menu item.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_ScreenShot (void)
{
	// NB: CPU thread should already be stopped at this point.

	char			fn[MAX_PATH] = {0};
	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	reply.lStructSize		= sizeof (OPENFILENAME);
	reply.hwndOwner 		= NULL; 	// LCD_GetWindow; -- Don't do this, or dialog will be too small!
	reply.hInstance 		= gInstance;
	reply.lpstrFilter		= "Bitmap Files (*.bmp)\0*.bmp\0";
	reply.nFilterIndex		= 1;
	reply.lpstrFile 		= fn;
	reply.nMaxFile			= sizeof (fn);
	reply.lpstrInitialDir	= NULL;
	reply.Flags 			= OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY | OFN_ENABLEHOOK | OFN_EXPLORER;
	reply.lpstrDefExt		= "bmp";
	reply.lpfnHook			= &::CenterDlgProc;

	if (::GetSaveFileName (&reply))
	{
		HANDLE	file = ::CreateFile (reply.lpstrFile, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
									FILE_ATTRIBUTE_NORMAL, NULL);

		if (file != INVALID_HANDLE_VALUE)
		{
			Screen::BufferInfo	info;
			
			if (Screen::GetBits (info, false /* !incremental */))
			{
				long	numColors = 1 << info.lcdDepth;

				DWORD	fileHeaderSize		= sizeof (BITMAPFILEHEADER);
				DWORD	bitmapHeaderSize	= sizeof (BITMAPINFOHEADER);
				DWORD	colorTableSize		= numColors * sizeof (RGBQUAD);
				DWORD	bitmapOffset		= fileHeaderSize + bitmapHeaderSize + colorTableSize;
				DWORD	bitmapSize			= info.myRowBytes * info.height;
				DWORD	fileSize			= bitmapOffset + bitmapSize;

				BITMAPFILEHEADER	fileHeader;

				fileHeader.bfType		= 'MB';
				fileHeader.bfSize		= fileSize;
				fileHeader.bfReserved1	= 0;
				fileHeader.bfReserved2	= 0;
				fileHeader.bfOffBits	= bitmapOffset;

				BITMAPINFOHEADER	bitmapHeader;

				bitmapHeader.biSize 			= sizeof (BITMAPINFOHEADER);
				bitmapHeader.biWidth			= info.myRowBytes * 8 / info.myDepth;
				bitmapHeader.biHeight			= info.height;
				bitmapHeader.biPlanes			= 1;
				bitmapHeader.biBitCount 		= info.myDepth;
				bitmapHeader.biCompression		= BI_RGB;
				bitmapHeader.biSizeImage		= 0;
				bitmapHeader.biXPelsPerMeter	= 0;
				bitmapHeader.biYPelsPerMeter	= 0;
				bitmapHeader.biClrUsed			= numColors;
				bitmapHeader.biClrImportant 	= 0;

				RGBQUAD*	colors = (RGBQUAD*) _alloca (colorTableSize);
				::PrvSetColorRamp (colors, numColors, info.myCLUT);

				DWORD	written;

				::WriteFile (file, &fileHeader, fileHeaderSize, &written, NULL);
				::WriteFile (file, &bitmapHeader, bitmapHeaderSize, &written, NULL);
				::WriteFile (file, colors, colorTableSize, &written, NULL);

				// What the...?  When I tried writing the negative of height to
				// indicate a top-down DIB, neither Paintbrush nor VC++ could
				// open the file!  So let's do it the hard way...

	//			WriteFile (file, bitmap, bitmapSize, &written, NULL);

				for (int yy = info.height - 1; yy >= 0; --yy)
				{
					void*	src = info.myBuffer + yy * info.myRowBytes;
					::WriteFile (file, src, info.myRowBytes, &written, NULL);
				}
			}

			::CloseHandle (file);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvLoadResourceBitmap
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	
 *
 * RETURNED:	
 *
 ***********************************************************************/

void PrvLoadResourceBitmap (HINSTANCE hInstance,
							BITMAPINFO** bitmap, RGBQUAD** colors, int* numColors,
							HRGN* region, RECT* bounds)
{
	*bitmap = NULL;
	*colors = NULL;
	*numColors = 0;
	*region = NULL;

	EmStream*	imageStream = ::SkinGetSkinStream ();

	if (imageStream)
	{
		// Turn the JPEG image into BMP format.

		*bitmap = ::JPEGToBMP (*imageStream);

		// Free up the resource info.

		delete imageStream;
		imageStream = NULL;
	}
	else
	{
		*bitmap = ::PrvGetDefaultBitmap ();
	}

	// Get the color information from the DIB.

	::PrvGetColors (*bitmap, ::PrvGetBits (*bitmap), colors, numColors);

	// Calculate a bounding region for the bitmap.

	*region = ::PrvCreateOutlineRegion (*bitmap);

	// Calculate the bounding rectangle for the bitmap.

	bounds->left = 0;
	bounds->top = 0;
	bounds->right = (*bitmap)->bmiHeader.biWidth;
	bounds->bottom = (*bitmap)->bmiHeader.biHeight;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetDefaultBitmap
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	
 *
 * RETURNED:	.
 *
 ***********************************************************************/

BITMAPINFO* PrvGetDefaultBitmap (void)
{
	BITMAPINFO*					result = NULL;

	UINT						bitmapID = 0;
	Preference<ScaleType>		scalePref (kPrefKeyScale);

	if (*scalePref == 1)
	{
		bitmapID = IDB_DEFAULT_SMALL;
	}
	else if (*scalePref == 2)
	{
		bitmapID = IDB_DEFAULT_LARGE;
	}
	else
	{
		assert (false);
	}

	if (bitmapID)
	{
		HRSRC	hResInfo = ::FindResource (gInstance, MAKEINTRESOURCE (bitmapID), RT_BITMAP);
		HGLOBAL	hBitmap = ::LoadResource (gInstance, hResInfo);
		LPVOID	pBitmap = ::LockResource (hBitmap);
		DWORD	bitmapSize = ::SizeofResource (gInstance, hResInfo);

		result = (BITMAPINFO*) Platform::AllocateMemory (bitmapSize);
		memcpy (result, pBitmap, bitmapSize);
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	PrvCreateDIBPalette
 *
 * DESCRIPTION:	Create and return a palette for the givem bitmap.
 *
 *				From KnowledgeBase article Q124947.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

HPALETTE PrvCreateDIBPalette (LPBITMAPINFO dib, void* bits)
{
	HPALETTE	hPal = NULL;

	if (::PrvIsPaletteDevice (NULL))
	{
		int numColors = ::PrvGetNumColors (dib);

		if (numColors)
		{
			LOGPALETTE* lpPal = (LOGPALETTE*) _alloca (sizeof (LOGPALETTE) + sizeof (PALETTEENTRY) * numColors);
			lpPal->palVersion		= 0x300;
			lpPal->palNumEntries	= numColors;

			for (int ii = 0; ii < numColors; ++ii)
			{
				lpPal->palPalEntry[ii].peRed	= dib->bmiColors[ii].rgbRed;
				lpPal->palPalEntry[ii].peGreen	= dib->bmiColors[ii].rgbGreen;
				lpPal->palPalEntry[ii].peBlue	= dib->bmiColors[ii].rgbBlue;
				lpPal->palPalEntry[ii].peFlags	= 0;
			}

			hPal = ::CreatePalette (lpPal);
		}
		else
		{
			// Figure out how many colors to use, maxing out at 256 - 20 - 16
			// (to allow for the 20 system palette colors, and reserving 16
			// colors for the LCD part of the display).

			const int	kNumMaxColors		= 256;
			const int	kNumSysColors		= 20;
			const int	kNumLCDColors		= 16;
			const int	kNumReqColors		= kNumMaxColors - kNumSysColors - kNumLCDColors;
			const int	kQuantizationLevel	= 6;

			CQuantizer q (kNumReqColors, kQuantizationLevel);
			q.ProcessImage (dib, bits);

			// Get the number of colors required for the image, and
			// get the actual colors.

			UINT		numQColors = q.GetColorCount ();
			RGBQUAD*	prgb = (RGBQUAD*) _alloca (numQColors * sizeof (RGBQUAD));
			q.GetColorTable (prgb);

			// Allocate memory for the palette.

			LOGPALETTE* lpPal = (LOGPALETTE*) _alloca (sizeof (LOGPALETTE) + sizeof (PALETTEENTRY) * numQColors);
			lpPal->palVersion		= 0x300;
			lpPal->palNumEntries	= numQColors;

			// Copy the entries to our palette.

			for (int ii = 0; ii < numQColors; ++ii)
			{
				lpPal->palPalEntry[ii].peRed	= prgb[ii].rgbRed;
				lpPal->palPalEntry[ii].peGreen	= prgb[ii].rgbGreen;
				lpPal->palPalEntry[ii].peBlue	= prgb[ii].rgbBlue;
				lpPal->palPalEntry[ii].peFlags	= 0;
			}

			hPal = ::CreatePalette (lpPal);
		}
	}

	return hPal;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetColors
 *
 * DESCRIPTION:	Determine the set of colors used by the given image.
 *				This set of colors is used to later generate a palette
 *				to be used when blitting.
 *
 * PARAMETERS:	dib - BITMAPINFO describing the image.
 *
 *				bits - pointer to the pixels of the image.
 *
 *				colors - pointer to the pointer to contain the array of
 *					colors that we'll want in our palette.	Storage for
 *					this array is allocated by this function, but the
 *					caller is responsible for later deleting it with
 *					Platform::DisposeMemory.
 *
 *				numColors - the number of entries in the array this
 *					function creates and returns in *colors.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void PrvGetColors (LPBITMAPINFO dib, void* bits, RGBQUAD** colors, int* numColors)
{
	int	numDIBColors = ::PrvGetNumColors(dib);
	if (numDIBColors != 0)
	{
		long	tableSize = numDIBColors * sizeof (RGBQUAD);
		*numColors = numDIBColors;
		*colors = (RGBQUAD*) Platform::AllocateMemory (tableSize);
		memcpy (*colors, dib->bmiColors, tableSize);
		return;
	}

	// Figure out how many colors to use, maxing out at 256 - 20
	// (to allow for the 20 system palette colors).

	const int	kNumMaxColors		= 256;
	const int	kNumSysColors		= 20;
	const int	kNumLCDColors		= 16;
	const int	kNumReqColors		= kNumMaxColors - kNumSysColors - kNumLCDColors;
	const int	kQuantizationLevel	= 6;

	CQuantizer q (kNumReqColors, kQuantizationLevel);
	q.ProcessImage (dib, bits);

	// Get the number of colors required for the image, and
	// get the actual colors.

	*numColors = q.GetColorCount ();
	*colors = (RGBQUAD*) Platform::AllocateMemory (*numColors * sizeof (RGBQUAD));
	q.GetColorTable (*colors);
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetPalette
 *
 * DESCRIPTION:	Return a palette to be used for all our drawing.  This
 *				palette includes entries for both the case graphics and
 *				the colors in the LCD area.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	The palette to be selected and realized.  This function
 *				owns the palette; the caller does not need to dispose
 *				of it.	Returns NULL if the screen is not a palette-
 *				based device.
 *
 ***********************************************************************/

HPALETTE PrvGetPalette (const Screen::BufferInfo& info)
{
	// If the screen device is not palette-based, don't return a palette.

	if (!::PrvIsPaletteDevice (NULL))
	{
		return NULL;
	}

	// If the screen device is palette-based, return a palette, creating
	// and caching it if necessary.

	if (!gCachedPalette)
	{
		int numLCDColors	= 1 << info.lcdDepth;
		int numTotalColors	= numLCDColors + gNumCaseColors;

		int lpalSize		= offsetof (LOGPALETTE, palPalEntry) + sizeof (PALETTEENTRY) * numTotalColors;

		// Allocate memory for the palette.

		LOGPALETTE* lpPal		= (LOGPALETTE*) _alloca (lpalSize);
		lpPal->palVersion		= 0x300;
		lpPal->palNumEntries	= numTotalColors;

		// Copy the entries to our palette.

		int palIndex = 0;
		int ii;

		// Copy in the colors for the LCD.

		for (ii = 0; ii < numLCDColors; ++ii, ++palIndex)
		{
			lpPal->palPalEntry[palIndex].peRed		= info.myCLUT[ii].fRed; 
			lpPal->palPalEntry[palIndex].peGreen	= info.myCLUT[ii].fGreen;
			lpPal->palPalEntry[palIndex].peBlue 	= info.myCLUT[ii].fBlue;
			lpPal->palPalEntry[palIndex].peFlags	= 0;
		}

		// Copy in the colors for the case.

		for (ii = 0; ii < gNumCaseColors; ++ii, ++palIndex)
		{
			lpPal->palPalEntry[palIndex].peRed		= gCaseColors[ii].rgbRed;
			lpPal->palPalEntry[palIndex].peGreen	= gCaseColors[ii].rgbGreen;
			lpPal->palPalEntry[palIndex].peBlue 	= gCaseColors[ii].rgbBlue;
			lpPal->palPalEntry[palIndex].peFlags	= 0;
		}

		gCachedPalette = ::CreatePalette (lpPal);
	}

	return gCachedPalette;
}



/***********************************************************************
 *
 * FUNCTION:	PrvIsPaletteDevice
 *
 * DESCRIPTION:	Returns whether or not the given device is palette-
 *				based or not.
 *
 * PARAMETERS:	hDC - the device to test.  If NULL, the screen device
 *					is tested.
 *
 * RETURNED:	True if the device is palette-based.  False otherwise.
 *
 ***********************************************************************/

BOOL PrvIsPaletteDevice (HDC hDC)
{
	int 	rasterCaps;

	if (hDC)
	{
		rasterCaps = ::GetDeviceCaps (hDC, RASTERCAPS);
	}
	else
	{
		hDC = ::GetDC (NULL);
		rasterCaps = ::GetDeviceCaps (hDC, RASTERCAPS);
		::ReleaseDC (NULL, hDC);
	}

	BOOL	isPaletteDevice = RC_PALETTE == (rasterCaps & RC_PALETTE);

	return isPaletteDevice;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetRowBytes
 *
 * DESCRIPTION:	Determine the number of bytes in a scanline, given the
 *				number of pixels in the line and assuming long-word
 *				rounding.
 *
 * PARAMETERS:	dib - BITMAPINFO describing the bitmap whose rowbytes
 *					we want to calculate.
 *
 *				width - the width in pixels of the image whose rowbytes
 *					we want to calculate.
 *
 *				bpp - the bit depth (bits per pixel) of the image whose
 *					rowbytes we want to calculate.
 *
 * RETURNED:	The number of bytes between the start of successive
 *				rows in the image.
 *
 ***********************************************************************/

long PrvGetRowBytes (LPBITMAPINFO dib)
{
	return PrvGetRowBytes (dib->bmiHeader.biWidth, dib->bmiHeader.biBitCount);
}


long PrvGetRowBytes (long width, long bpp)
{
	return ((width * bpp + 31) & ~31) / 8;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetNumColors
 *
 * DESCRIPTION:	Determine the number of colors in the color table for
 *				the given BITMAPINFO.  The number of colors is
 *				determined by the depth of the bitmap, and is even
 *				optional if the bit depth is greater than 8 bpp.
 *
 * PARAMETERS:	dib - a pointer to the BITMAPINFO containing the
 *					bitmap and color information.
 *
 * RETURNED:	For 1-, 4-, and 8-bpp images returns the number of
 *				colors in the contained color table, as determined by
 *				the image's bit depth and the biClrUsed field.	For
 *				images with more bpp, there is no color table, so
 *				this function returns zero.
 *
 ***********************************************************************/

int PrvGetNumColors (LPBITMAPINFO dib)
{
	int numColors;

	if (dib->bmiHeader.biBitCount <= 8)
		numColors = (1 << dib->bmiHeader.biBitCount);
	else
		numColors = 0;	// No palette needed for 24 BPP DIB

	if (dib->bmiHeader.biClrUsed > 0)
		numColors = dib->bmiHeader.biClrUsed;  // Use biClrUsed

	return numColors;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetBits
 *
 * DESCRIPTION:	Calculate the location of the pixels of a BITMAPINFO
 *				object.
 *
 * PARAMETERS:	dib - the pointer to the BITMAPINFO describing the
 *					bitmap.  This function assumes that the pixels
 *					are stored in memory right after the variable-
 *					length (and optional) color table.
 *
 * RETURNED:	A pointer to the pixels.
 *
 ***********************************************************************/

BYTE* PrvGetBits (LPBITMAPINFO dib)
{
	return ((BYTE*) dib) + dib->bmiHeader.biSize + ::PrvGetNumColors (dib) * sizeof(RGBQUAD);
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetRECT
 *
 * DESCRIPTION:	Convert a Palm OS RectangleType to a Windows RECT.
 *
 * PARAMETERS:	r - the Palm OS RectangleType to convert
 *
 * RETURNED:	The rectangle in RECT format.
 *
 ***********************************************************************/

RECT PrvGetRECT (const RectangleType& r)
{
	RECT	result;

	result.left 	= r.topLeft.x;
	result.top		= r.topLeft.y;
	result.right	= r.topLeft.x + r.extent.x;
	result.bottom	= r.topLeft.y + r.extent.y;

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	PrvCreateOutlineRegion
 *
 * DESCRIPTION:	Create an outline region for the given bitmap.	It is
 *				assumed that the bitmap is in 24-bit mode, and that
 *				the areas around the image to clip out are white.
 *
 * PARAMETERS:	dib - information about the bitmap.
 *
 * RETURNED:	Handle to the region surrounding the bitmap.
 *
 ***********************************************************************/

HRGN PrvCreateOutlineRegion (BITMAPINFO* dib)
{
	HRGN	region = NULL;

	int 	width = dib->bmiHeader.biWidth;
	int 	height = dib->bmiHeader.biHeight;
	int 	bitmapSize = height * ::PrvGetRowBytes (width, 1);

	// Convert the image to 1-bpp.

	void*	bitmap = _alloca (bitmapSize);
	if (::PrvGetNumColors (dib) == 0)
		::PrvConvert24To1BPP (bitmap, ::PrvGetBits (dib), width, height);
	else
		::PrvConvert8To1BPP (bitmap, ::PrvGetBits (dib), width, height, dib->bmiColors);

	// Create a mask from the 1-bpp image using a seedfill.

	void*	mask = _alloca (bitmapSize);
	::PrvMakeMask (mask, bitmap, width, height);

	// Create the region from the 1-bpp image.

	region = ::PrvBitmapToRegion (mask, width, height);

	// Clean up and exit.
	return region;
}


/***********************************************************************
 *
 * FUNCTION:	PrvConvert24To1BPP
 *
 * DESCRIPTION:	Convert the given 24-bpp image to a 1-bpp image
 *
 * PARAMETERS:	dstPtr - buffer to hold the 1-bpp image.  The caller is
 *					responsible for making sure it's large enough.
 *
 *				srcPtr - buffer containing the 24-bpp image.
 *
 *				width - width of the image in pixels.
 *
 *				height - height of the image in pixels.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void PrvConvert24To1BPP (void* dstPtr, void* srcPtr, int width, int height)
{
	long	srcRowBytes 	= ::PrvGetRowBytes (width, 24);
	long	destRowBytes	= ::PrvGetRowBytes (width, 1);

	for (int yy = 0; yy < height; ++yy)
	{
		uint8*	src = ((uint8*) srcPtr) + yy * srcRowBytes;
		uint8*	dest = ((uint8*) dstPtr) + yy * destRowBytes;

		uint8	bitMask = 0x80;
		uint8	aByte	= 0;

		for (int xx = 0; xx < width; ++xx)
		{
			// See if the current pixel is dark.  Test it this way
			// to ensure that all three src++'s are evaluated.

			Bool	isDark1 = *src++ < 0xC0;
			Bool	isDark2 = *src++ < 0xC0;
			Bool	isDark3 = *src++ < 0xC0;

			if (isDark1 || isDark2 || isDark3)
			{
				aByte |= bitMask;
			}

			bitMask >>= 1;
			if (bitMask == 0)
			{
				*dest++ = aByte;
				bitMask = 0x80;
				aByte = 0;
			}
		}

		if (bitMask != 0)
		{
			*dest++ = aByte;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvConvert8To1BPP
 *
 * DESCRIPTION:	Convert the given 8-bpp image to a 1-bpp image
 *
 * PARAMETERS:	dstPtr - buffer to hold the 1-bpp image.  The caller is
 *					responsible for making sure it's large enough.
 *
 *				srcPtr - buffer containing the 8-bpp image.
 *
 *				width - width of the image in pixels.
 *
 *				height - height of the image in pixels.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void PrvConvert8To1BPP (void* dstPtr, void* srcPtr, int width, int height, RGBQUAD* colors)
{
	long	srcRowBytes 	= ::PrvGetRowBytes (width, 8);
	long	destRowBytes	= ::PrvGetRowBytes (width, 1);

	for (int yy = 0; yy < height; ++yy)
	{
		uint8*	src = ((uint8*) srcPtr) + yy * srcRowBytes;
		uint8*	dest = ((uint8*) dstPtr) + yy * destRowBytes;

		uint8	bitMask = 0x80;
		uint8	aByte	= 0;

		for (int xx = 0; xx < width; ++xx)
		{
			// See if the current pixel is dark.  Test it this way
			// to ensure that all three src++'s are evaluated.

			uint8	pixel = *src++;
			RGBQUAD	color = colors[pixel];

			Bool	isDark1 = color.rgbRed < 0xC0;
			Bool	isDark2 = color.rgbGreen < 0xC0;
			Bool	isDark3 = color.rgbBlue < 0xC0;

			if (isDark1 || isDark2 || isDark3)
			{
				aByte |= bitMask;
			}

			bitMask >>= 1;
			if (bitMask == 0)
			{
				*dest++ = aByte;
				bitMask = 0x80;
				aByte = 0;
			}
		}

		if (bitMask != 0)
		{
			*dest++ = aByte;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvMakeMask
 *
 * DESCRIPTION:	Calculate a mask for the given 1-bpp image.
 *
 * PARAMETERS:	dstPtr - buffer to hold the generated mask.  Caller is
 *					responsible for making the buffer is large enough.
 *
 *				srcPtr - pointer to the input 1-bpp image.
 *
 *				width - width of the source image, in pixels.
 *
 *				height - height of the source image, in pixels.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void PrvMakeMask (	void* dstPtr, void* srcPtr, int width, int height)
{
	uint8	*dest, *src, *prev, *prev0;
	uint8	srcVal, destVal, prevDestVal, newDestVal;
	int 	rowBytes = ::PrvGetRowBytes (width, 1);
	int 	absRowBytes = ::PrvGetRowBytes (width, 1);
	int 	xx, yy;
	int 	valChanged, anyChange, pass;

	uint8	edge = 0;	// Sentry for the left and right edges.

	//
	// Set the destination bitmap to black (1's)
	//
	dest = (uint8*) dstPtr;
	yy = height;
	do {
		xx = rowBytes;
		do {
			*dest++ = (uint8) ~0;
		} while (--xx != 0);
	} while (--yy != 0);

	//
	// Allocate a scanline buffer on the stack.  This scanline
	// acts as a sentry at the top and bottom of the dest buffer.
	//
	prev0 = (uint8*) _alloca (rowBytes);

	xx = rowBytes;
	dest = prev0;
	do {
		*dest++ = edge;
	} while (--xx != 0);

	src = (uint8*) srcPtr;
	dest = (uint8*) dstPtr;
	pass = 0;

	do {	// Loop while we're still effectively smearing.
		++pass;
		prev = prev0;
		anyChange = false;
		yy = height;

		do {	// Loop through all the scanlines.
			//
			// Smear down and to the right.
			//
			prevDestVal = edge;
			xx = absRowBytes;

			do {	// Loop through all the bytes in this line.
				destVal = *dest;

				if (destVal != 0)	// If it's not already all-white, let's see if we can mess with it.
				{
					srcVal = *src;

					// The following code effectively pushes down any 0's from
					// the previous line, then adds any 1's from the source.
					destVal = (destVal & *prev) | srcVal;

					do {
						// Shift destVal to the right.	The upper bit of the
						// new value gets filled in with the lower bit of the
						// previous destVal.
						newDestVal = (destVal >> 1) | (prevDestVal << 7);

						// AND the shifted destVal with the pre-shifted destVal.
						// This has the effect of smearing any 0's to the right.
						newDestVal &= destVal;

						// OR in the corresponding byte from the source.  This
						// has the effect of stopping the smear when it hits a 1.
						newDestVal |= srcVal;

						// See if the destVal we're working on has changed.  If
						// so, keep smearing to the right.
						if (valChanged = (newDestVal != destVal))
						{
							destVal = newDestVal;
						}
					} while (valChanged);

					// If this destVal has changed, record the fact and write
					// out the new value.
					if (*dest != destVal)
					{
						anyChange = true;
						*dest = destVal;
					}
				}

				// Move to the next bytes in this line.
				++src;
				++prev;
				++dest;
				prevDestVal = destVal;
			} while (--xx != 0);

			//
			// Smear down and to the left.	Same logic as before.
			//
			prevDestVal = edge;
			xx = absRowBytes;
			do {
				--src;
				--prev;
				--dest;
				destVal = *dest;

				if (destVal != 0)
				{
					srcVal = *src;
					destVal = (destVal & *prev) | srcVal;

					do {
						newDestVal = (destVal << 1) | (prevDestVal >> 7);
						newDestVal &= destVal;
						newDestVal |= srcVal;
						if (valChanged = (newDestVal != destVal))
							destVal = newDestVal;
					} while (valChanged);

					if (*dest != destVal)
					{
						anyChange = true;
						*dest = destVal;
					}
				}

				prevDestVal = destVal;
			} while (--xx != 0);

			// Move to the next scanlines.	Our movement is either
			// up or down, depending on whether rowBytes is negative
			// or not.
			src += rowBytes;
			prev = dest;
			dest += rowBytes;
		} while (--yy != 0);

		//
		// Switch direction, adjust ptrs, and loop until no change.
		// Make sure we do at least one up and one down pass, though.
		//
		rowBytes = -rowBytes;
		src += rowBytes;
		dest += rowBytes;
	} while (anyChange || (pass < 2));
}


/***********************************************************************
 *
 * FUNCTION:	PrvBitmapToRegion
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

HRGN PrvBitmapToRegion (void* srcPtr, int width, int height)
{
	HRGN	region		= NULL;

	uint8*	bits		= (uint8*) srcPtr;
	long	rowBytes	= ::PrvGetRowBytes (width, 1);

	for (long yy = 0; yy < height; ++yy)
	{
		Bool	wasInRegion = false;
		long	xStart;
		long	xEnd;

		uint8* cBits = bits + ((height - yy) - 1) * rowBytes;

		uint8	aByte;
		uint8	bitMask = 0;

		for (long xx = 0; xx < width; ++xx, bitMask >>= 1)
		{
			if (bitMask == 0)
			{
				bitMask = 0x80;
				aByte = *cBits++;
			}

			Bool	nowInRegion = (aByte & bitMask) != 0;

			// Entering a region?
			if (nowInRegion && !wasInRegion)
			{
				// Start new "in region" scan
				xStart = xx;
			}

			// Exiting a region?
			else if (!nowInRegion && wasInRegion)
			{
				// Close off "in region" scan.
				xEnd = xx;
				::PrvAddToRegion (region, yy, xStart, xEnd);
			}

			wasInRegion = nowInRegion;
		}

		// End of scan line. Check to see if the region scan is still
		// open.  If so, close it.

		if (wasInRegion)
		{
			xEnd = width;
			::PrvAddToRegion (region, yy, xStart, xEnd);
		}
	}

	return region;
}


/***********************************************************************
 *
 * FUNCTION:	PrvAddToRegion
 *
 * DESCRIPTION:	Add a rectangle to the given rection.  If the region
 *				has not yet been allocated, allocate it.
 *
 * PARAMETERS:	region - region to create or modify.
 *
 *				top - top coordinate of the rectangle to add.  The bottom
 *					of the rectangle is yy + 1.
 *
 *				left - left coordinate of the rectangle to add.
 *
 *				right - right coordinate of the rectangle to add.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void PrvAddToRegion (HRGN& region, int top, int left, int right)
{
	if (region == NULL)
	{
		region = ::CreateRectRgn (left, top, right, top + 1);
	}
	else
	{
		HRGN	tempRgn = ::CreateRectRgn (left, top, right, top + 1);
		::CombineRgn (region, region, tempRgn, RGN_OR);
		::DeleteObject (tempRgn);
	}
}


void	PrvGetLCDBounds	(RectangleType& result)
{
	int	index = 0;
	SkinElementType	type;
	RectangleType	bounds;

	while (::SkinGetElementInfo (index, type, bounds))
	{
		if (type == kElement_LCD)
		{
			result = bounds;
			return;
		}

		++index;
	}

	assert (false);
}


void	PrvGetLCDBounds	(AbsRectType& bounds)
{
	RectangleType	r;
	::PrvGetLCDBounds (r);

	bounds.left		= r.topLeft.x;
	bounds.top		= r.topLeft.y;
	bounds.right	= r.topLeft.x + r.extent.x;
	bounds.bottom	= r.topLeft.y + r.extent.y;
}
