/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Platform_FfsLib.h"

#include "Emulator.h"			// gInstance
#include "Logging.h"			// LogAppendMsg
#include "Marshal.h"
#include "Platform.h"			// AllocateMemory
#include "UAE_Utils.h"

#include <direct.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>


#define MAX_FFS_PATH		256

static char					gHome[MAX_FFS_PATH + 1];
static int					gHomeLength;
static char					gOrgHome[MAX_FFS_PATH + 1];
static ffs_ffblk			gFindBlk;
static long					gFindHandle;
static struct _finddata_t	gFindData;
static int					gSavedErrno;
static UInt16				gSavedAttrib;


#define PRINTF	if (1) ; else LogAppendMsg


/************************************************************
 *  FUNCTION    : ActualPath
 *
 *  DESCRIPTION : This function maps an FFS path to the actual path on the
 *                user's hard drive.
 *
 *                Example : If the user specifies that he wants c:\tmp to
 *                          be the directory to emulate a CF card.  A call of
 *                          ActualPath("\\mike") will return "c:\\tmp\\mike".
 *              
 *  PARAMETERS  : path - Ffs path
 *              
 *  RETURNS     : Actual path on user's hard drive.
 *************************************************************/

static char* ActualPath (char* path)
{
	static char actual_path[MAX_FFS_PATH * 2 + 1];

	if (path[0] == '\\')
	{
		strcpy (actual_path, gOrgHome);
		strcat (actual_path, path + 1);
	}
	else
	{
		strcpy (actual_path, gHome);
		strcat (actual_path, path);
	}

	return actual_path;
}


/************************************************************
 *  FUNCTION    : XferFindData
 *
 *  DESCRIPTION : Microsoft does not implement FindFirst(), FindNext()
 *                exactly like FFS.  This function copies the return info
 *                from Microsoft's Find data into FFS's find data.
 *              
 *  PARAMETERS  : gFindBlk  - FFS find data structure
 *                gFindData - Microsoft find data structure.
 *              
 *  RETURNS     : Actual path on user's hard drive.
 *************************************************************/

static void XferFindData (ffs_ffblk* gFindBlk, struct _finddata_t* gFindData)
{
	static char	tmp[MAX_FFS_PATH + 1];
	static char	filename[13];
	static char	ext[13];

	gFindBlk->ff_attrib	= (char) gFindData->attrib;
	gFindBlk->ff_fsize	= gFindData->size;

	strcpy (gFindBlk->ff_longname, gFindData->name);
	GetShortPathName (ActualPath (gFindData->name), tmp, MAX_FFS_PATH);

	_splitpath (tmp, NULL, NULL, filename, ext);

/*---------------------------------------------------------------------------
 * . and .. are not handled properly by _splitpath().
 *--------------------------------------------------------------------------*/
	if (!strcmp (gFindData->name, ".") || !strcmp (gFindData->name, ".."))
	{
		strcpy (gFindBlk->ff_name, gFindData->name);
	}
	else
	{
		strcpy (gFindBlk->ff_name, filename);
		strcat (gFindBlk->ff_name, ext);
	}
}


/************************************************************
 *  FUNCTION    : TranslateModeToPlatform() & TranslateModeToFfs()
 *
 *  DESCRIPTION : Not all compilers agree on the mode bit definitions.
 *                These to functions translate between FFs's definition 
 *                and the platform's compiler defition.
 *              
 *  PARAMETERS  : 
 *  RETURNS     : 
 *************************************************************/

static int TranslateModeToPlatform (int mode)
{
	int retval = 0;

	if (mode & FFS_S_IREAD)
		retval |= _S_IREAD;
	if (mode & FFS_S_IWRITE)
		retval |= _S_IWRITE;
	if (mode & FFS_S_IFCHR)
		retval |= _S_IFCHR;
	if (mode & FFS_S_IFDIR)
		retval |= _S_IFDIR;
	if (mode & FFS_S_IFREG)
		retval |= _S_IFREG;

	return retval;
}

static int TranslateModeToFfs (int mode)
{
	int retval = 0;

	if (mode & _S_IREAD)
		retval |= FFS_S_IREAD;
	if (mode & _S_IWRITE)
		retval |= FFS_S_IWRITE;
	if (mode & _S_IFCHR)
		retval |= FFS_S_IFCHR;
	if (mode & _S_IFDIR)
		retval |= FFS_S_IFDIR;
	if (mode & _S_IFREG)
		retval |= FFS_S_IFREG;

	return retval;
}

//--------------------------------------------------
// Library initialization, shutdown, sleep and wake
//--------------------------------------------------

Err Platform_FfsLib::LibOpen (UInt16 libRefNum)
{
	UNUSED_PARAM(libRefNum)


	// Determine the home directory for the FFS system.  The home directory
	// is stored in the preference FfsHome.  It can have the following values:
	//
	//		non-existant:	use Poser's directory
	//		empty:			use Poser's directory
	//		"foo":			use <PoserPath>\foo
	//		"foo\bar":		use <PoserPath>\foo\bar
	//		"\foo\bar":		use <PoserDrive>\foo\bar
	//		"X:\foo\bar":	use X:\foo\bar
	//
	// If the directory doesn't exist, Poser will try to create it.  If the
	// attempt to create the directory fails, Poser will use Poser's directory.

	// Default the home directory to the Poser directory.

	gHome[0] = 0;
	::GetModuleFileName (gInstance, gHome, MAX_FFS_PATH);

	// Get pieces for it in case we need them.

	char	pdrive[_MAX_DRIVE], pdir[_MAX_DIR];
	_splitpath (gHome, pdrive, pdir, NULL, NULL);
	_makepath (gHome, pdrive, pdir, NULL, NULL);

	// See if the user's specified a home directory.

	Preference<string>	ffsHomePref (kPrefKeyFfsHome);
	if (ffsHomePref.Loaded ())
	{
		string	ffsHome = *ffsHomePref;
		if (!ffsHome.empty ())
		{
			if (ffsHome[ffsHome.size() - 1] != '\\')
				ffsHome += '\\';

			// If so, get its pieces to see if we need to provide any.

			char	path[_MAX_PATH], drive[_MAX_DRIVE], dir[_MAX_DIR], fileName[_MAX_FNAME];
			_splitpath (ffsHome.c_str (), drive, dir, fileName, NULL);

			// If no drive, use Poser's drive.

			if (drive[0] == 0)
				strcpy (drive, pdrive);

			// If no directory, use Poser's directory

			if (dir[0] == 0)
				strcpy (dir, pdir);

			// else if a relative directory, prepend Poser's directory

			else if (dir[0] != '\\')
			{
				char	temp[_MAX_DIR];
				strcpy (temp, dir);
				strcpy (dir, pdir);
				strcat (dir, temp);
			}


			// Build it all back up.

			_makepath (gHome, drive, dir, fileName, NULL);
		}
	}

	// If the directory doesn't exist, make it

	DWORD	attr = GetFileAttributes (gHome);
	if (attr == -1)
	{
		if (CreateDirectory (gHome, NULL) == 0)
		{
			// Couldn't make the directory, use the default.
			::GetModuleFileName (gInstance, gHome, MAX_FFS_PATH);
			_splitpath (gHome, pdrive, pdir, NULL, NULL);
			_makepath (gHome, pdrive, pdir, NULL, NULL);
		}
	}

	if (gHome[strlen (gHome) - 1] != '\\')
		strcat (gHome, "\\");

	strcpy (gOrgHome, gHome);

	gHomeLength = strlen (gHome);

	return errNone;
}

Err Platform_FfsLib::LibClose (UInt16 libRefNum)
{
	UNUSED_PARAM(libRefNum)
	return errNone;
}


Err Platform_FfsLib::LibSleep (UInt16 libRefNum)
{
	UNUSED_PARAM(libRefNum)
	return errNone;
}


Err Platform_FfsLib::LibWake (UInt16 libRefNum)
{
	UNUSED_PARAM(libRefNum)
	return errNone;
}


/************************************************************
 *  FUNCTION    : FfsChangeOccurred()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Boolean Platform_FfsLib::FfsChangeOccurred (UInt16 libRef)
{
	return FALSE;
}


/************************************************************
 *  FUNCTION    : FfsCardIsATA()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Boolean Platform_FfsLib::FfsCardIsATA (UInt16 libRef, UInt8 drive_num)
{
	return TRUE;
}


/************************************************************
 *  FUNCTION    : FfsCardIsInserted()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Boolean Platform_FfsLib::FfsCardIsInserted (UInt16 libRef, UInt8 drive_num)
{
	return TRUE;
}


/************************************************************
 *  FUNCTION    : FfsSetDebuggingOn()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

void Platform_FfsLib::FfsSetDebuggingOn (UInt16 libRef, UInt16 s_port, UInt16 level)
{
/*---------------------------------------------------------------------------
 * Not implemented
 *--------------------------------------------------------------------------*/
}


/************************************************************
 *  FUNCTION    : FfsSetDebuggingOff()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

void Platform_FfsLib::FfsSetDebuggingOff (UInt16 libRef)
{
/*---------------------------------------------------------------------------
 * Not implemented
 *--------------------------------------------------------------------------*/
}


/************************************************************
 *  FUNCTION    : FfsUnInstallErrorHandler()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

void Platform_FfsLib::FfsUnInstallErrorHandler (UInt16 libRef)
{
/*---------------------------------------------------------------------------
 * Not implemented
 *--------------------------------------------------------------------------*/
}


/************************************************************
 *  FUNCTION    : FfsInstallErrorHandler()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

void Platform_FfsLib::FfsInstallErrorHandler (UInt16 libRef, Int16 (*CritErr)(Int16, Int16, char*))
{
/*---------------------------------------------------------------------------
 * Not implemented
 *--------------------------------------------------------------------------*/
}


/************************************************************
 *  FUNCTION    : FfsCreat()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Int16 Platform_FfsLib::FfsCreat (UInt16 libRef, char* path, Int16 mode)
{
	Int16 retval;

	mode		= TranslateModeToPlatform (mode);
	retval		= (Int16) _creat (ActualPath (path), mode);
	gSavedErrno	= errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsEof()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsEof (UInt16 libRef, Int16 handle)
{
	Err retval;

	retval = (Err) _eof ((int) handle);
	gSavedErrno = errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsUnlink()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsUnlink (UInt16 refNum, char* path)
{
	Err retval;

	retval = (Err) _unlink (ActualPath (path));
	gSavedErrno = errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsSetdrive()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

void Platform_FfsLib::FfsSetdrive (UInt16 libRef, UInt16 drive, UInt16* ndrives)
{
	*ndrives = 1;
}


/************************************************************
 *  FUNCTION    : FfsGetdrive()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

void Platform_FfsLib::FfsGetdrive (UInt16 libRef, UInt16* drive)
{
	*drive = 1;
}


/************************************************************
 *  FUNCTION    : FfsTell()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Int32 Platform_FfsLib::FfsTell (UInt16 libRef, Int16 handle)
{
	Int32 retval;

	retval = _tell ((int) handle);
	gSavedErrno = errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsGetErrno()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Int16 Platform_FfsLib::FfsGetErrno (UInt16 libRef)
{
	return (Int16) gSavedErrno;
}


/************************************************************
 *  FUNCTION    : FfsLseek()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Int32 Platform_FfsLib::FfsLseek (UInt16 libRef, Int16 handle, Int32 offset, Int16 origin)
{
	Int32 retval;

	retval = (Int32) _lseek (handle, offset, origin);
	gSavedErrno = errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsIsDir()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsIsDir (UInt16 libRef, char* path, Boolean* is_dir)
{
	Err				retval;
	struct _stat	lstat;

	retval = _stat (ActualPath (path), &lstat);

/*---------------------------------------------------------------------------
 * Temp fix for bug in library.
 *--------------------------------------------------------------------------*/
	if ((retval != 0) && (errno == 2))
	{
		errno		= 0;
		gSavedErrno	= errno;
		*is_dir		= FALSE;

		return 0;
	}

	gSavedErrno = errno;

	if (lstat.st_mode & _S_IFDIR)
		*is_dir = TRUE;
	else
		*is_dir = FALSE;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsFstat()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsFstat (UInt16 libRef, Int16 handle, ffs_stat* pstat)
{
	Err				retval;
	struct _stat	lstat;

	if ((retval = (Err) _fstat ((int) handle, &lstat)) != 0)
		return(retval);

	gSavedErrno = errno;

	pstat->st_dev	= lstat.st_dev;
	pstat->st_mode	= TranslateModeToFfs (lstat.st_mode);
	pstat->st_nlink	= 1;
	pstat->st_rdev	= lstat.st_rdev;
	pstat->st_size	= lstat.st_size;
	pstat->st_atime	= *(ffs_date_t*) &lstat.st_atime;
	pstat->st_mtime	= *(ffs_date_t*) &lstat.st_mtime;
	pstat->st_ctime	= *(ffs_date_t*) &lstat.st_ctime;

/*---------------------------------------------------------------------------
 * pstat->st_attr Not implemented
 *--------------------------------------------------------------------------*/

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsStat()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsStat (UInt16 libRef, char* path, ffs_stat* pstat)
{
	Err				retval;
	struct _stat	lstat;

	retval			= _stat (ActualPath (path), &lstat);

	gSavedErrno		= errno;

	pstat->st_dev	= lstat.st_dev;
	pstat->st_mode	= TranslateModeToFfs (lstat.st_mode);
	pstat->st_nlink	= 1;
	pstat->st_rdev	= lstat.st_rdev;
	pstat->st_size	= lstat.st_size;
	pstat->st_atime	= *(ffs_date_t*) &lstat.st_atime;
	pstat->st_mtime	= *(ffs_date_t*) &lstat.st_mtime;
	pstat->st_ctime	= *(ffs_date_t*) &lstat.st_ctime;
	pstat->st_attr	= (UInt8) GetFileAttributes (ActualPath (path)) & 0xFF;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsFlushDisk()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsFlushDisk (UInt16 libRef, UInt16 drive)
{
/*---------------------------------------------------------------------------
 * Not implemented
 *--------------------------------------------------------------------------*/
	return errNone;
}


/************************************************************
 *  FUNCTION    : FfsFlush()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsFlush (UInt16 libRef, Int16 handle)
{
	Err retval;

	retval = (Err) _commit ((int) handle);
	gSavedErrno = errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsSetfileattr()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsSetfileattr (UInt16 libRef, char* name, UInt16 attr)
{
	if (!SetFileAttributes(ActualPath(name), attr))
	{
		gSavedErrno = errno;
		return(-1);
	}

	gSavedErrno = errno;
	return errNone;
}


/************************************************************
 *  FUNCTION    : FfsGetfileattr()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsGetfileattr (UInt16 libRef, char* name, UInt16* attr)
{
	*attr		= (UInt16) (GetFileAttributes (ActualPath (name)) & 0xFF);
	gSavedErrno	= errno;

	return errNone;
}


/************************************************************
 *  FUNCTION    : FfsRename()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsRename (UInt16 refNum, char* path, char* new_name)
{
	Err retval;

	retval = (Err) rename (ActualPath (path), new_name);
	gSavedErrno = errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsFormat()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsFormat (UInt16 refNum, UInt16 drive)
{
/*---------------------------------------------------------------------------
 * Never to implement, too dangerous.
 *--------------------------------------------------------------------------*/
	gSavedErrno = -99;
	return -1;
}


/************************************************************
 *  FUNCTION    : FfsRmdir()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsRmdir (UInt16 refNum, char* dirname)
{
	Err retval;

	retval = (Err) _rmdir (ActualPath (dirname));
	gSavedErrno = errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsMkdir()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsMkdir (UInt16 refNum, char* dirname)
{
	Err retval;

	retval = (Err) _mkdir (ActualPath (dirname));
	gSavedErrno = errno;

	return(retval);
}


/************************************************************
 *  FUNCTION    : FfsGetcwd()
 *                
 *  DESCRIPTION :  Note: Had to change path to "void *" in order to get
 *                       MACROS in TrapPatches_Ffslib.cpp
 *              
 *                       Need to put in check to make sure buflen is big
 *                       enough to hold entire path.
 *
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

char* Platform_FfsLib::FfsGetcwd (UInt16 refNum, char* path, Int16 buflen)
{
	char* retval;

	if (path == NULL)
		retval = (char*) Platform::AllocateMemory (strlen (gHome + (gHomeLength - 1) + 1));
	else
		retval = path;

	strcpy (retval, gHome + (gHomeLength - 1));

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsChdir()
 *
 *  DESCRIPTION : Note: If the user pics a subdirectory on their hard drive
 *                      to emulate a CF memory card, it is possible to chdir
 *                      "above" the root of the drive.  This function  does
 *                      not check for that condition. 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsChdir (UInt16 refNum, char* path)
{
	if (_fullpath (gHome, ActualPath( path), MAX_FFS_PATH) == NULL)
	{
		gSavedErrno = errno;
		return(-1);
	}

	if (gHome[strlen (gHome)-1] != '\\')
		strcat (gHome, "\\");

	gSavedErrno = errno;

	return errNone;
}


/************************************************************
 *  FUNCTION    : FfsRemove()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsRemove (UInt16 refNum, char* path)
{
	Err retval;

	retval = (Err) remove (ActualPath (path));
	gSavedErrno = errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsWrite()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Int16 Platform_FfsLib::FfsWrite (UInt16 refNum, Int16 handle, void* buffer, Int16 num_bytes)
{
	Int16 retval;

	retval = (Int16) _write (handle, buffer, (UInt16) num_bytes);
	gSavedErrno = errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsRead()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Int16 Platform_FfsLib::FfsRead (UInt16 refNum, Int16 handle, void* buffer, Int16 num_bytes)
{
	Int16 retval;

	retval		= _read(handle, buffer, (UInt16) num_bytes);
	gSavedErrno	= errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsOpen()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Int16 Platform_FfsLib::FfsOpen (UInt16 refNum, char* path, Int16 flags, Int16 mode)
{
	Int16 retval;

/*---------------------------------------------------------------------------
 * Notes: Ffs library only supports O_BINARY and always opend in that mode.
 *--------------------------------------------------------------------------*/
	mode		= TranslateModeToPlatform (mode);
	retval		= (Int16) _open (ActualPath (path), (int) flags | _O_BINARY, (int) mode);
	gSavedErrno	= errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsClose()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsClose (UInt16 refNum, Int16 handle)
{
	Err retval;

	retval = (Err) _close(handle);
	gSavedErrno = errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsFindfirst()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsFindfirst (UInt16 refNum, char* path, Int16 attrib, ffs_ffblk* ff_blk)
{
	UInt8	attr_match;

	PRINTF ("FindFirst(%s,0x%04X)", path, attrib);

	if ((gFindHandle = _findfirst (ActualPath (path), &gFindData)) == -1)
	{
		gSavedErrno = errno;
		return -1;
	}

	gSavedAttrib	= attrib;
	attr_match		= (UInt8) (attrib & 0x00FF);

/*---------------------------------------------------------------------------
 * search until we find one with matching attribute
 *--------------------------------------------------------------------------*/
	for(;;)
	{
/*---------------------------------------------------------------------------
 * Make sure attributes match, ignore the archive attribute for now.
 *--------------------------------------------------------------------------*/
		if ((UInt16) attrib == FA_ALL)
			break;

		if ((gFindData.attrib & ~FA_ARCH) == attr_match)
			break;

/*---------------------------------------------------------------------------
 * get the next entry
 *--------------------------------------------------------------------------*/
		if (_findnext (gFindHandle, &gFindData) == -1)
		{
			gSavedErrno = errno;
			return -1;
		}
	}

	XferFindData (ff_blk, &gFindData);
	gSavedErrno = errno;

	return errNone;
}


/************************************************************
 *  FUNCTION    : FfsFindnext()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsFindnext (UInt16 refNum, ffs_ffblk* ff_blk)
{
	UInt8	attr_match;

	attr_match = (UInt8) (gSavedAttrib & 0x00FF);

/*---------------------------------------------------------------------------
 * search until we find one with matching attribute
 *--------------------------------------------------------------------------*/
	for(;;)
	{
		if (_findnext (gFindHandle, &gFindData) == -1)
		{
			gSavedErrno = errno;
			return -1;
		}

/*---------------------------------------------------------------------------
 * Make sure attributes match, ignore the archive attribute for now.
 *--------------------------------------------------------------------------*/
		if (gSavedAttrib == FA_ALL)
			break;

		if ((gFindData.attrib & ~FA_ARCH) == attr_match)
			break;

	}

	XferFindData(ff_blk, &gFindData);
	gSavedErrno = errno;

	return errNone;
}


/************************************************************
 *  FUNCTION    : FfsFinddone()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsFinddone (UInt16 refNum, ffs_ffblk* ff_blk)
{
	Err retval;

	retval = (Err)_findclose(gFindHandle);
	gSavedErrno = errno;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsGetdiskfree()
 *
 *  DESCRIPTION : 
 *              
 *  PARAMETERS  :
 *              
 *  RETURNS     :
 *************************************************************/

Err Platform_FfsLib::FfsGetdiskfree (UInt16 libRef, UInt8 drive_num, ffs_diskfree_t* dtable)
{
	Err			retval;
	_diskfree_t	disk_free;

	drive_num	= (toupper (gHome[0]) - 'A') + 1;
	retval		= (Err) _getdiskfree (drive_num, &disk_free);
	gSavedErrno	= errno;

	dtable->total_clusters		= disk_free.total_clusters;
	dtable->avail_clusters		= disk_free.avail_clusters;
	dtable->sectors_per_cluster	= disk_free.sectors_per_cluster;
	dtable->bytes_per_sector	= disk_free.bytes_per_sector;

	return retval;
}


/************************************************************
 *  FUNCTION    : FfsGetLibAPIVersion()
 *
 *  DESCRIPTION :    
 *      Get our liFfsGetLibAPIVersion()brary API version.  The Ffs library does not
 *      need to be "opened" to call FfsGetLibAPIVersion.
 *
 *  PARAMETERS  : refNum      -- Ffs library reference number
 *                dwVerP      -- pointer to variable for storing the version number
 *
 *  RETURNS     : 0           -- no error
 *************************************************************/

Err Platform_FfsLib::FfsGetLibAPIVersion (UInt16 refNum, UInt32* dwVerP)
{
	*dwVerP = 0x01023000;
	return errNone;
}
